 
from ..config import print_specialized_function_names
from ..ndtypes.closure_type import _closure_type_cache

def clear_specializations():   
  for clos_t in _closure_type_cache.itervalues():
    clos_t.specializations.clear()


def print_specializations():

  if print_specialized_function_names:
    print
    print "FUNCTION SPECIALIZATIONS"
    count = 0
    for ((untyped,closure_types), clos_t) in sorted(_closure_type_cache.items()):
      specializations = clos_t.specializations.items()
      if len(specializations) > 0:
        name = untyped.name if hasattr(untyped, 'name') else str(untyped)
        print "Closure %s %s" % (name, closure_types)
        for (arg_types, typed_fn) in sorted(specializations):
          print "  -- %s ==> %s" % (arg_types, typed_fn.name)
          count += 1
    print
    print "Total: %d function specializations" % count


import atexit
atexit.register(print_specializations)


from parakeet.testing_helpers import expect, run_local_tests  


def covariance(x,y):
  return ((x-x.mean()) * (y-y.mean())).mean()

def fit_simple_regression(x,y):
  slope = covariance(x,y) / covariance(x,x)
  offset = y.mean() - slope * x.mean() 
  return slope, offset

import numpy as np 

def test_simple_regression():
  N = 10
  x = np.random.randn(N).astype('float64')
  slope = 903.29
  offset = 102.1
  y = slope * x + offset
  expect(fit_simple_regression, [x,y], (slope,offset))


if __name__ == '__main__':
  run_local_tests()






import numpy as np 
from parakeet import testing_helpers

def add_2d(x):
  return x[1:-1, 2:] + x[1:-1, :-2] - 4 * x[1:-1, 1:-1]

def test_add_2d():
  x = np.array([[1,2,3,4,5,6],[10,20,30,40,50,60]])
  testing_helpers.expect(add_2d, [x], add_2d(x))

if __name__ == "__main__":
  testing_helpers.run_local_tests()




import parakeet
from parakeet.testing_helpers import run_local_tests, expect_eq

def mk_adder(t):
  f, b, (x,y) = parakeet.build_fn([t,t], t)
  b.return_(b.add(x,y))
  return f

def double_i64(x):
  fn = mk_adder(parakeet.Int64)
  return parakeet.run_typed_fn(fn, (x,x))
  
def double_f64(x):
  fn = mk_adder(parakeet.Float64)
  return parakeet.run_typed_fn(fn, (x,x))

def test_add():
  expect_eq(double_i64(1), 2)
  expect_eq(double_i64(-1), -2)
  expect_eq(double_f64(1.0), 2.0)
  expect_eq(double_f64(-1.0), -2.0)


if __name__ == '__main__':
  run_local_tests()  




# Rosenbrock function derivative
# 
# Copied from https://github.com/numfocus/python-benchmarks/blob/master/rosen_der/rosen_der_python.py
# Original authors: Travis Oliphant (NumPy version) & Serge Guelton (loops version)
#

import numpy as np 
from parakeet import jit 

@jit 
def rosenbrock_derivative(x):
  der = np.empty_like(x)
  der[1:-1] = (+ 200 * (x[1:-1] - x[:-2] ** 2)
               - 400 * (x[2:] - x[1:-1] ** 2) * x[1:-1]
               - 2 * (1 - x[1:-1]))
  der[0] = -400 * x[0] * (x[1] - x[0] ** 2) - 2 * (1 - x[0])
  der[-1] = 200 * (x[-1] - x[-2] ** 2)
  return der

N = 12
x = np.arange(N) / float(N)
print "Input: ", x
print "Deriv(Rosenbrock):", rosenbrock_derivative(x)




######################################
#          BACKEND SELECTION         #
######################################
#
#  'c': sequential, use gcc or clang to compile
#  'openmp': multi-threaded execution for array operations, requires gcc 4.4+
#  'llvm': deprecated
#  'interp': interpreter, will be dreadfully slow
#  'cuda': experimental GPU support
#

backend = 'openmp' 

######################################
#        PARAKEET OPTIMIZATIONS      #
######################################
  
    
opt_inline = True

opt_fusion = True
opt_combine_nested_maps = True

opt_specialize_fn_args = True 

# experimental!
opt_simplify_array_operators = False

opt_index_elimination = True
opt_range_propagation = True

opt_licm = True
opt_redundant_load_elimination = True
opt_stack_allocation = True

opt_shape_elim = True 

# replace 
#   a = alloc
#   ...
#   b[i:j] = a
#
#   with 
#     a = b[i:j]  
opt_copy_elimination = True

# may dramatically increase compile time
opt_loop_unrolling = False

# suspiciously complex optimizations may introduce bugs 
# TODO: comb through carefully 
opt_scalar_replacement = False
    
# run verifier after each transformation 
opt_verify = False

# recompile functions for distinct patterns of unit strides and 0 or 1 input values 
value_specialization = True 



#####################################
#            DEBUG OUTPUT           #
#####################################

# show untyped IR after it's translated from Python?
print_untyped_function = False

# show the higher level typed function after specialization?
print_specialized_function = False 

# show function after all data adverbs like Map/Reduce/Scan have been 
# lowered to use indexing explicitly into their inputs 
print_indexified_function = False

# print function after all adverbs have been turned to loops
print_loopy_function = False

# show lower level typed function before
# it gets translated to LLVM?
print_lowered_function = False

# before starting function specialization, print the fn name and input types 
print_before_specialization = False

# show the input function to each transformation?
print_functions_before_transforms =  []
                                        
# show the function produced by each transformation?
print_functions_after_transforms =   []

# show aliases and escape sets
print_escape_analysis = False

# how long did each transform take?
print_transform_timings = False

# print each transform's name when it runs
print_transform_names = False

# at exit, print the names of all specialized functions
print_specialized_function_names = False

# tell the backend to print whatever code it generates, 
# whether it's C, CUDA, or LLVM 
print_generated_code = False 

#####################################
#         DESPERATE MEASURES        #
#####################################

testing_find_broken_transform = False 




from ast_conversion import translate_function_value, translate_function_ast
from closure_specializations import print_specializations
from decorators import jit, macro, staged_macro, typed_macro, axis_macro
from diagnose import find_broken_transform
from run_function import run_untyped_fn, run_typed_fn, run_python_fn, specialize
import type_conv_decls as _decls 
from typed_repr import typed_repr




from .. ndtypes import (Int8, Int16, Int32, Int64, 
                        UInt8, UInt16, UInt32, UInt64,
                        Float32, Float64, Bool) 

from .. frontend import macro, jit 
from .. syntax import Cast 

@macro 
def int8(x):
  return Cast(x, type = Int8) 

@macro 
def int16(x):
  return Cast(x, type = Int16) 

@macro 
def int32(x):
  return Cast(x, type = Int32) 

@macro 
def int64(x):
  return Cast(x, type = Int64) 


@macro 
def uint8(x):
  return Cast(x, type = UInt8) 

@macro 
def uint16(x):
  return Cast(x, type = UInt16) 

@macro 
def uint32(x):
  return Cast(x, type = UInt32) 

@macro 
def uint64(x):
  return Cast(x, type = UInt64)

uint = uint64 

@macro 
def float32(x):
  return Cast(x, type = Float32)

@macro 
def float64(x):
  return Cast(x, type = Float64)

@macro 
def bool(x):
  return Cast(x, type = Bool)



import ctypes 

from llvm.ee import GenericValue 

from .. ndtypes import FloatT, SignedT, IntT, NoneT, PtrT    
from .. ndtypes import type_conv 
import llvm_types 

def python_to_generic_value(x, t):
  if isinstance(t, FloatT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.real(llvm_t, x)
  elif isinstance(t, SignedT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.int_signed(llvm_t, x)
  elif isinstance(t, IntT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.int(llvm_t, x)
  elif isinstance(t, PtrT):
    return GenericValue.pointer(x)
  else:
    ctypes_obj = type_conv.from_python(x)
    return GenericValue.pointer(ctypes.addressof(ctypes_obj))

def ctypes_to_generic_value(cval, t):
  if isinstance(t, FloatT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.real(llvm_t, cval.value)
  elif isinstance(t, SignedT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.int_signed(llvm_t, cval.value)
  elif isinstance(t, IntT):
    llvm_t = llvm_types.llvm_value_type(t)
    return GenericValue.int(llvm_t, cval.value)
  elif isinstance(t, NoneT):
    return GenericValue.int(llvm_types.int64_t, 0)
  elif isinstance(t, PtrT):
    return GenericValue.pointer(ctypes.addressof(cval.contents))
  else:
    return GenericValue.pointer(ctypes.addressof(cval))

def generic_value_to_python(gv, t):
  if isinstance(t, SignedT):
    return t.dtype.type(gv.as_int_signed() )
  elif isinstance(t, IntT):
    return t.dtype.type( gv.as_int() )
  elif isinstance(t, FloatT):
    llvm_t = llvm_types.ctypes_scalar_to_lltype(t.ctypes_repr)
    return t.dtype.type(gv.as_real(llvm_t))
  elif isinstance(t, NoneT):
    return None
  else:
    addr = gv.as_pointer()
    struct = t.ctypes_repr.from_address(addr)
    return t.to_python(struct)


import multiprocessing 

from ..syntax import Expr, Tuple
from ..syntax.helpers import get_fn, return_type
from ..ndtypes import ScalarT, TupleT, ArrayT
from ..c_backend import PyModuleCompiler

import config 

class MulticoreCompiler(PyModuleCompiler):
  
  def __init__(self, depth = 0, *args, **kwargs):
    self.depth = depth
    self.seen_parfor = None 
    PyModuleCompiler.__init__(self, *args, **kwargs)
  
  @property 
  def cache_key(self):
    return self.__class__, self.depth > 0
  
  _loop_var_names = ["i","j","k","l","a","b","c","ii","jj","kk","ll","aa","bb","cc"] 
  def loop_vars(self, count, init_value = "0"):
    assert count <= len(self._loop_var_names)
    return [self.fresh_var("int64_t", self._loop_var_names[i], init_value) 
            for i in xrange(count)]
       
  def visit_NumCores(self, expr):
    # by default we're running sequentially 
    return "%d" % multiprocessing.cpu_count()
  
  def tuple_to_var_list(self, expr):
    assert isinstance(expr, Expr)
    if isinstance(expr, Tuple):
      return self.visit_expr_list(expr)
    elif isinstance(expr.type, TupleT):
      c_val = self.visit_expr(expr)
      return ["%s.elt%d" % (c_val, i) for i in xrange(len(expr.type.elt_types))]
    else:
      assert isinstance(expr.type, ScalarT), "Unexpected expr %s : %s" % (expr, expr.type)
      return [self.visit_expr(expr)]
  
  
  def get_fn_name(self, fn_expr, attributes = [], inline = True):
    return PyModuleCompiler.get_fn_name(self, fn_expr, 
                                        compiler_kwargs = {'depth':self.depth}, 
                                        attributes = attributes, 
                                        inline = inline)
  
  def get_fn_info(self, fn_expr, attributes = [], inline = True):
    """
    Given a function expression, return:
      - the name of its C representation
      - C expressions representing its closure arguments
      - Parakeet input types 
    """

    fn_name = self.get_fn_name(fn_expr, attributes = attributes, inline = inline)
    closure_args = self.get_closure_args(fn_expr)
    root_fn = get_fn(fn_expr)
    input_types = root_fn.input_types 
    return fn_name, closure_args, input_types
    
  
  def build_loop_body(self, fn_expr, loop_vars, target_name = None):
    """
    Inside a loop nest, construct an index tuple, call the given function, 
    and optionally assign its result to the target variable. 
    Returns the string representation of the loop body and the set of
    private variables it uses. 
    """
    fn_name, closure_args, input_types = self.get_fn_info(fn_expr)

    private_vars = [loop_var for loop_var in loop_vars] 
    last_input_type = input_types[-1]
    body = ""
    if isinstance(last_input_type, TupleT):
      # make a tupke out of the loop variables
      
      c_tuple_t = self.to_ctype(last_input_type)
      index_tuple = self.fresh_var(c_tuple_t, "index_tuple")
      private_vars.append(index_tuple)
      for i, loop_var in enumerate(loop_vars):
        body += "\n%s.elt%d = %s;\n" % (index_tuple, i, loop_var)
      combined_args = tuple(closure_args) + (index_tuple,)
      call = "%s(%s)" % (fn_name, ", ".join(combined_args))
    else:
      combined_args = tuple(closure_args) + tuple(loop_vars)
      call = "%s(%s)" % (fn_name, ", ".join(combined_args))
    if target_name:
      body += "%s = %s;\n" % (target_name, call)
    else:
      body += "\n%s;\n" % call
    return body, private_vars
  
  
  def enter_parfor(self):
    self.depth += 1
    if not self.seen_parfor:
      self.seen_parfor = True 
      self.add_compile_flag("-fopenmp")
      self.add_link_flag("-fopenmp")
    
  def exit_parfor(self):
    self.depth -= 1

  def visit_ParFor(self, stmt):
    bounds = self.tuple_to_var_list(stmt.bounds)
    n_vars = len(bounds)
    loop_vars = self.loop_vars(n_vars)
    
    self.enter_parfor()
    body, private_vars = self.build_loop_body(stmt.fn, loop_vars)
    loops = self.build_loops(loop_vars, bounds, body)
    self.exit_parfor()
    
    if self.depth == 0:  
      release_gil = "\nPy_BEGIN_ALLOW_THREADS\n"
      acquire_gil = "\nPy_END_ALLOW_THREADS\n" 
      
      
      if config.collapse_nested_loops:
        omp = "#pragma omp parallel for private(%s) schedule(%s)" % \
          (", ".join(private_vars), config.schedule)
        if len(loop_vars) > 1:
          omp += " collapse(%d)" % len(loop_vars)
      else:
        omp = "#pragma omp parallel for private(%s) schedule(%s)" % \
          (private_vars[0], config.schedule)
      return release_gil + omp + loops + acquire_gil    
    else:
      return loops 
     
  def visit_IndexReduce(self, expr):
    """
    For now, just use a sequential implementation for reductions
    """
    bounds = self.tuple_to_var_list(expr.shape)
    n_vars = len(bounds)
    combine_name, combine_closure_args, _ = self.get_fn_info(expr.combine)
    loop_vars = self.loop_vars(n_vars)
    assert expr.init is not None, "Accumulator required but not given"
    
    
    elt = self.fresh_var(return_type(expr.fn), "elt")

    body, _ = self.build_loop_body(expr.fn, loop_vars, target_name = elt)
    acc = self.fresh_var(expr.type, "acc", self.visit_expr(expr.init))
    combine_arg_str = ", ".join(tuple(combine_closure_args) + (acc, elt))
    body += "\n%s = %s(%s);\n" % (acc, combine_name, combine_arg_str)
    self.append(self.build_loops(loop_vars, bounds, body))
    return acc 
    
  def visit_IndexScan(self, expr):
    """
    For now, just use a sequential implementation for scans
    """
    assert isinstance(expr.type, ArrayT), "Expected output of Scan to be an array"
    
    bounds = self.tuple_to_var_list(expr.shape)
    n_vars = len(bounds)
    
    combine_name, combine_closure_args, _ = self.get_fn_info(expr.combine)
    loop_vars = self.loop_vars(n_vars)
    
    
    result = self.alloc_array(expr.type, expr.shape)
    
    assert expr.init is not None, "Accumulator required but not given"
    
    elt_t = return_type(expr.fn) 
    assert isinstance(elt_t, ScalarT), "Scans of non-scalar values (%s) not yet implemented" % elt_t
    elt = self.fresh_var(elt_t, "elt")
    body, _ = self.build_loop_body(expr.fn, loop_vars, target_name = elt)
    acc = self.fresh_var(expr.init.type, "acc", self.visit_expr(expr.init))
    combine_arg_str = ", ".join(tuple(combine_closure_args) + (acc, elt))
    body += "\n%s = %s(%s);\n" % (acc, combine_name, combine_arg_str)
    emit_name, emit_closure_args, _ = self.get_fn_info(expr.emit)
    body += "\n"
    emit_args = tuple(emit_closure_args) + (acc,)
    emit_args_str = ", ".join(emit_args)
    body += self.setidx(result, 
                        loop_vars, 
                        "%s(%s)" % (emit_name, emit_args_str), 
                        full_array = True, 
                        return_stmt = True)
    self.append(self.build_loops(loop_vars, bounds, body))
    return result
    
  def visit_Map(self, expr):
    assert False, "Map should have been lowered into ParFor by now: %s" % expr 
  
  def visit_OuterMap(self, expr):
    assert False, "OuterMap should have been lowered into ParFor by now: %s" % expr 
  
  def visit_Reduce(self, expr):
    assert False, "Reduce should have been lowered into ParFor by now: %s" % expr 
  
  def visit_Scan(self, expr):
    assert False, "Scan should have been lowered into ParFor by now: %s" % expr
    
  def visit_IndexMap(self, expr):
    assert False, "IndexMap should have been lowered into ParFor by now: %s" % expr 
  



import numpy as np 
from .. import prims 
from ..syntax import Const, PrimCall, Var, If, Select 
from ..syntax.helpers import (zero, one, 
                              is_zero, is_one, const_bool, 
                              wrap_constants, get_types, wrap_if_constant)

from core_builder import CoreBuilder 

class ArithBuilder(CoreBuilder):
  def prim(self, prim_fn, args, name=None):
    args = wrap_constants(args)
    arg_types = get_types(args)
    upcast_types = prim_fn.expected_input_types(arg_types)
    result_type = prim_fn.result_type(upcast_types)
    upcast_args = [self.cast(x, t) for (x,t) in zip(args, upcast_types)]
    prim_call = PrimCall(prim_fn, upcast_args, type=result_type)
    if name:
      return self.assign_name(prim_call, name)
    else:
      return prim_call

  def pick_first(self, x, y):
    """Return x but first cast it to the common type of both args"""

    return self.cast(x, x.type.combine(y.type))

  def pick_second(self, x, y):
    """Return y but first cast it to the common type of both args"""

    return self.cast(y, x.type.combine(y.type))

  def pick_const(self, x, y, c):
    """Return a constant cast to the common type of both args"""

    return self.cast(wrap_if_constant(c), x.type.combine(y.type))

  def add(self, x, y, name = None):
    if is_zero(x):
      return self.pick_second(x,y)
    elif is_zero(y):
      return self.pick_first(x,y)
    elif x.__class__ is Const and y.__class__ is Const:
      return self.pick_const(x, y, x.value + y.value)
    if name is None: name = "add_result"
    return self.prim(prims.add, [x,y], name)

  def sub(self, x, y, name = None):
    if is_zero(y):
      return self.pick_first(x,y)
    elif x.__class__ is Const and y.__class__ is Const:
      return self.pick_const(x, y, x.value - y.value)
    
    if name is None: name = "sub_result"
    return self.prim(prims.subtract, [x,y], name)

  def mul(self, x, y, name = None):
    if is_one(x):
      return self.pick_second(x,y)
    elif is_one(y):
      return self.pick_first(x,y)
    elif is_zero(x) or is_zero(y):
      return self.pick_const(x, y, 0)
    
    if name is None: name = "mul_result"
    return self.prim(prims.multiply, [x,y], name)

  def div(self, x, y, name = None):
    if is_one(y):
      return self.pick_first(x,y)
    elif x.__class__ is Const and y.__class__ is Const:
      return self.pick_const(x, y, x.value / y.value)
    else:
      return self.prim(prims.divide, [x,y], name)


  
  
  def mod(self, x, y, name = None):
    if is_one(y):
      return self.pick_const(x, y, 0)
    elif x.__class__ is Const and y.__class__ is Const:
      return self.pick_const(x, y, x.value % y.value)
    if name is None: name = "mod_result"
    return self.prim(prims.mod, [x,y], name)

  def fmod(self, x, y, name = None):
    if is_one(y):
      return self.pick_const(x, y, 0)
    elif x.__class__ is Const and y.__class__ is Const:
      return self.pick_const(x, y, np.fmod(x,y))
    else:
      return self.prim(prims.fmod, [x,y], name)
      
  def div_round_up(self, x, y, name = None):
    div_toward_zero = self.div(x,y,"div_toward_zero")
    m = self.fmod(x, y)
    has_remainder = self.neq(m, zero(m.type), "has_remainder") 
    prod = self.mul(x, y, "prod")
    prod_is_positive = self.is_positive(prod, "prod_is_positive")
    was_rounded_down = self.and_(has_remainder, prod_is_positive)
    div_plus_one = self.add(div_toward_zero, one(div_toward_zero.type), "div_plus_one")
    return self.select(was_rounded_down, div_plus_one, div_toward_zero, name = name)
  
  def safediv(self, x, y, name = None):
    """
    Like div_round_up but only safe for positive integers
    """
    top = self.add(x, y)
    top = self.sub(top, one(top.type))
    return self.div(top, y, name = name)
  
  
  def ceil(self, x, name = None):
    return self.prim(prims.ceil, [x], name)
  
  def floor(self, x, name = None):
    return self.prim(prims.floor, [x], name)
  
  def round(self, x, name = None):
    return self.prim(prims.round, [x], name)
  
  def lt(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return const_bool(False)
    if name is None: name = "lt_result"
    return self.prim(prims.less, [x,y], name)

  def lte(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return const_bool(True)
    
    if name is None: name = "lte_result"
    return self.prim(prims.less_equal, [x,y], name)

  
  def gt(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return const_bool(False)
    if name is None: name = "gt_result"
    return self.prim(prims.greater, [x,y], name)

  def is_positive(self, x, name = "is_positive"):
    return self.gt(x, zero(x.type), name)
  
  def gte(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return const_bool(True)
    if name is None: name = "gte_result"
    return self.prim(prims.greater_equal, [x,y], name)

  def eq(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return const_bool(True)
    if name is None: name = "eq_result"
    return self.prim(prims.equal, [x,y], name)

  def neq(self, x, y, name = None):
    if isinstance(x, (Var, Const)) and x == y:
      return const_bool(False)
    if name is None: name = "neq_result"
    return self.prim(prims.not_equal, [x,y], name)

  def min(self, x, y, name = None):
    assert x.type == y.type, \
        "Type mismatch between %s and %s" % (x, y)
    if x.__class__ is Const and y.__class__ is Const:
      return x if x.value < y.value else y 
    cond = self.lte(x, y)
    expr = Select(cond, x, y, type = x.type)
    if name is None: return expr 
    else: return self.assign_name(expr, name)
    
  def max(self, x, y, name = None):
    assert x.type == y.type, \
        "Type mismatch between %s and %s" % (x, y)
    if x.__class__ is Const and y.__class__ is Const:
      return x if x.value > y.value else y 
    cond = self.gte(x, y)
    expr = Select(cond, x, y, type = x.type)
    if name is None: return expr 
    else: self.assign_name(expr, name)
    
  def or_(self, x, y, name = None):
    if x.__class__ is Const and x.value:
      return x
    elif y.__class__ is Const and y.value:
      return y
    if name is None:
      name = "or_result"
    return self.prim(prims.logical_or, [x,y], name)
  
    
  def and_(self, x, y, name = None):
    if x.__class__ is Const and x.value and y.__class__ is Const and y.value:
      return x 
    if name is None:
      name = "and_result"
    return self.prim(prims.logical_and, [x,y], name)



from .. import syntax 
from ..ndtypes import ScalarT, TupleT 
from ..syntax import Expr, While, ForLoop, Const 
from ..syntax.helpers import zero, one, zero_i32, zero_i64, wrap_if_constant

from array_builder import ArrayBuilder 

class LoopBuilder(ArrayBuilder):
  
  """
  Builder for loops and things that use loops
  """
  
  
  
  def loop_var(self, name = "i", start_val = syntax.zero_i64):
    """
    Generate three SSA variables to use as the before/during/after values
    of a loop counter throughout some loop.

    By default initialize the counter to zero, but optionally start at different
    values using the 'start_val' keyword.
    """

    start_val = wrap_if_constant(start_val)
    counter_type = start_val.type

    counter_before = self.assign_name(start_val, name + "_before")
    counter = self.fresh_var(counter_type, name)
    counter_after = self.fresh_var(counter_type, name + "_after")
    merge = {counter.name:(counter_before, counter_after)}
    return counter, counter_after, merge

  def loop(self, 
           start, 
           niters, 
           loop_body,
           return_stmt = False,
           while_loop = False,
           step = None, 
           merge = None):
    
    if isinstance(start, (int,long)):
      start = self.int(start)
      
    assert isinstance(start, Expr)
    
    if isinstance(niters, (int,long)):
      niters = self.int(niters)
      
    assert isinstance(niters, Expr)
    
    if step is None:
      step = one(start.type)
    elif isinstance(step, (int, long)):
      step = self.int(step)
    
    assert isinstance(step, Expr)
    
    if merge is None:
      merge = {}
      
    if while_loop:
      i, i_after, i_merge = self.loop_var("i", start)
      merge.update(i_merge)
      cond = self.lt(i, niters)
      self.blocks.push()
      loop_body(i)
      self.assign(i_after, self.add(i, step))
      body = self.blocks.pop()
      loop_stmt = While(cond, body, merge)
    else:
      var = self.fresh_var(start.type, "i")
      self.blocks.push()
      loop_body(var)
      body = self.blocks.pop()
      loop_stmt = ForLoop(var, start, niters, step, body, merge)

    if return_stmt:
      return loop_stmt
    else:
      self.blocks += loop_stmt


  class Accumulator:
    def __init__(self, acc_type, fresh_var, assign):
      self.acc_type = acc_type
      self.fresh_var = fresh_var
      self.assign = assign
      self.start_var = fresh_var(acc_type, "acc")
      self.curr_var = self.start_var

    def get(self):
      return self.curr_var

    def update(self, new_value):
      new_var = self.fresh_var(self.acc_type, "acc")
      self.assign(new_var, new_value)
      self.curr_var = new_var

  def mk_acc(self, init):
    return self.Accumulator(init.type, self.fresh_var, self.assign)
    
  def accumulate_loop(self, start, stop, loop_body, init, return_stmt = False):
    acc = self.mk_acc(init)
    def loop_body_with_acc(i):
      loop_body(acc, i)
    loop_stmt = self.loop(start, stop, loop_body_with_acc, return_stmt = True)
    loop_stmt.merge[acc.start_var.name] = (init, acc.curr_var)
    if return_stmt:
      return loop_stmt, acc.start_var
    else:
      self.blocks += loop_stmt
      return acc.start_var 
    
  def array_copy(self, src, dest, return_stmt = False):
    assert self.is_array(dest)
    shape = self.shape(dest)
    dims = self.tuple_elts(shape)
    rank = len(dims)
    index_vars = []
    def create_loops():
      i = len(index_vars)
      def loop_body(index_var):
        index_vars.append(index_var)
        if i+1 == rank:
          index_tuple = self.tuple(index_vars, "idx")
          lhs = self.index(dest, index_tuple, temp=False)
          rhs = self.index(src, index_tuple, temp=True)
          self.assign(lhs, rhs)
        else:
          self.blocks += create_loops()
      start = syntax.zero_i64
      stop = dims[i]
      if i > 0 or return_stmt:
        return self.loop(start, stop, loop_body, True)
      else:
        return self.loop(start, stop, loop_body, return_stmt)

    return create_loops()
  
  def _to_list(self, bounds):
    if isinstance(bounds, Expr):
      if isinstance(bounds.type, ScalarT):
        return [bounds]
      else:
        assert isinstance(bounds.type, TupleT), \
          "Expected tuple but got %s : %s" % (bounds, bounds.type) 
        return self.tuple_elts(bounds)
    elif isinstance(bounds, (int,long)):
      return [bounds]
    else:
      assert isinstance(bounds, (tuple,list))
      return bounds 
  
  
  def nested_loops(self, 
                     upper_bounds, 
                     loop_body, 
                     lower_bounds = None, 
                     step_sizes = None, 
                     index_vars_as_list = False):
    upper_bounds = self._to_list(upper_bounds)
    
    n_loops = len(upper_bounds)
    assert lower_bounds is None or len(lower_bounds) == n_loops
    assert step_sizes is None or len(step_sizes) == n_loops 
    
    if lower_bounds is None:
      lower_bounds = [self.int(0) for _ in upper_bounds]
    else:
      lower_bounds = self._to_list(lower_bounds)
    
    if step_sizes is None:
      step_sizes = [self.int(1) for _ in upper_bounds]
    else:
      step_sizes = self._to_list(step_sizes)
    
    
    def build_loops(index_vars = ()):
      n_indices = len(index_vars)
      if n_indices == n_loops:
        if isinstance(loop_body, Expr):
          input_types = self.input_types(loop_body)
          if len(input_types) == len(index_vars):
            result = self.call(loop_body, index_vars)
          else:
            result = self.call(loop_body, [self.tuple(index_vars)])
        else:
          if index_vars_as_list:
            idx_tuple = list(index_vars)
          elif n_indices > 1:
            idx_tuple = self.tuple(index_vars)
          else:
            idx_tuple = index_vars[0]
          assert hasattr(loop_body, '__call__'), "Expected callable value, got %s" % (loop_body,)
          result = loop_body(idx_tuple)
        assert self.is_none(result), "Expected loop body to return None, not %s" % (result,)
      else:
        def inner_loop_body(idx):
          build_loops(index_vars + (idx,))
        lower = lower_bounds[n_indices]
        upper = upper_bounds[n_indices]  
        step = step_sizes[n_indices]
        # sanity check the bounds 
        if lower.__class__ is Const and upper.__class__ is Const and step.__class__ is Const:
          if step.value < 0:
            assert lower.value > upper.value, \
              "Attempting to build invalid loop from %s to %s by %s" % (lower, upper, step)
          elif step.value > 0:
            assert upper.value > lower.value, \
              "Attempting to build invalid loop from %s to %s by %s" % (lower, upper, step) 
        self.loop(lower, upper, inner_loop_body, step=step)
    build_loops()


import parakeet 
from  parakeet import testing_helpers
import numpy as np 

@parakeet.jit
def setidx(x, idx, y):
    x[idx] = y
    return x

def test_set_1d_simple_slice():
    x = np.array([1,2,3,4,5,6]) 
    idx = slice(2,4)
    y = [10, 20]
    x2 = x.copy()
    x[idx] = y
    testing_helpers.expect(setidx, [x2, idx, y], x)
    
def test_set_1d_simple_slice_to_const():
    x = np.array([1,2,3,4,5,6]) 
    idx = slice(2,4)
    y = 0
    x2 = x.copy()
    x[idx] = y
    testing_helpers.expect(setidx, [x2, idx, y], x)

def test_set_1d_step_slice_to_const():
    x = np.array([1,2,3,4,5,6]) 
    idx = slice(2,4,2)
    y = 0
    x2 = x.copy()
    x[idx] = y
    testing_helpers.expect(setidx, (x2, idx, y), x)

def test_set_1d_negative_slice():
    x = np.array([1,2,3,4,5,6]) 
    idx = slice(4,2,-1)
    y = [10, 20]
    x2 = x.copy()
    x[idx] = y
    testing_helpers.expect(setidx, (x2, idx, y), x)

if __name__ == '__main__':
    testing_helpers.run_local_tests()




import os 
import subprocess
import time

import config 

class CommandFailed(Exception):
  def __init__(self, cmd, env, label):
    self.cmd = cmd 
    self.env = env 
    self.label = label 
    
    
  def __str__(self):
    return "CommandFailed(%s)" % self.cmd 
  
  def __repr__(self):
    return "CommandFailed(cmd=%s, env=%s, label=%s)" % (" ".join(self.cmd), self.env, self.label)
  
def run_cmd(cmd, env = None, label = ""):
  if config.print_commands: 
    print " ".join(cmd)
  if config.print_command_elapsed_time: 
    t = time.time()
  
  # first compile silently
  # if you encounter an error, then recompile with output printing
  try:
    if config.suppress_compiler_output: 
      with open(os.devnull, "w") as fnull:
        subprocess.check_call(cmd, stdout = fnull, stderr = fnull, env = env)
    else:
      subprocess.check_call(cmd, env = env)
  except:
    raise CommandFailed(cmd, env, label)
    
  if config.print_command_elapsed_time: 
    if label:
      print "%s, elapsed time: %0.4f" % (label, time.time() - t)
    else:
      print "Elapsed time:", time.time() - t 



import numpy as np 
from .. frontend import jit 
from adverbs import imap 

@jit
def pmap1(f, x, w = 3):
  n = x.shape[0]
  def local_apply(i):
    lower = __builtins__.max(i-w/2, 0)
    upper = __builtins__.min(i+w/2+1, n)
    elts = x[lower:upper]
    return f(elts)
  return imap(local_apply, n)

@jit  
def pmap2(f, x, width = (3,3)):
  """
  Patch-map where the function can accept both interior windows
  and smaller border windows
  """
  width_x, width_y = width 
  n_rows, n_cols = x.shape
  hx = width_x / 2
  hy = width_y / 2
  
  def local_apply((i,j)):
    lx = __builtins__.max(i-hx, 0)
    ux = __builtins__.min(i+hx+1, n_rows)
    ly = __builtins__.max(j-hy, 0)
    uy = __builtins__.min(j+hy+1, n_cols)
    return f(x[lx:ux, ly:uy])
    
  return imap(local_apply, x.shape)
    
@jit  
def pmap2_trim(f, x, width = (3,3), step = (1,1)):
  """
  Patch-map over interior windows, ignoring the border
  """
  width_x, width_y = width 
  step_x, step_y = step 
  n_rows, n_cols = x.shape
  hx = width_x / 2
  hy = width_y / 2
  return [[f(x[i-hx:i+hx+1, j-hy:j+hy+1]) 
           for j in np.arange(hx, n_cols-hx, step_x)] 
           for i in np.arange(hy, n_rows-hy, step_y)]






from .. import names 
from .. analysis.collect_vars import  collect_binding_names
from .. syntax import Var, Assign, ForLoop
from clone_function import CloneFunction
from transform import Transform

class CloneStmt(CloneFunction):
  def __init__(self, outer_type_env):
    Transform.__init__(self)
    self.recursive = False
    self.type_env = outer_type_env
    self.rename_dict = {}

  def rename(self, old_name):
    old_type = self.type_env[old_name]
    new_name = names.refresh(old_name)
    new_var = Var(new_name, old_type)
    self.rename_dict[old_name] = new_var
    self.type_env[new_name] = old_type
    return new_name

  def rename_var(self, old_var):
    new_name = names.refresh(old_var.name)
    new_var = Var(new_name, old_var.type)
    self.rename_dict[old_var.name] = new_var
    self.type_env[new_name] = old_var.type
    return new_var

  def transform_merge(self, merge):
    new_merge = {}
    for (old_name, (l,r)) in merge.iteritems():
      new_name = self.rename(old_name)
      new_left = self.transform_expr(l)
      new_right = self.transform_expr(r)
      new_merge[new_name] = (new_left, new_right)
    return new_merge

  def transform_merge_before_loop(self, merge):
    new_merge = {}
    for (old_name, (l,r)) in merge.iteritems():
      new_name = self.rename(old_name)

      new_left = self.transform_expr(l)
      new_merge[new_name] = (new_left, r)
    return new_merge

  def transform_merge_after_loop(self, merge):
    for (new_name, (new_left, old_right)) in merge.items():
      merge[new_name] = (new_left, self.transform_expr(old_right))
    return merge

  def transform_Assign(self, expr):
    for name in collect_binding_names(expr.lhs):
      self.rename(name)
    new_lhs = self.transform_expr(expr.lhs)
    new_rhs = self.transform_expr(expr.rhs)
    return Assign(new_lhs, new_rhs)

  def transform_Var(self, expr):
    return self.rename_dict.get(expr.name, expr)

  def transform_ForLoop(self, stmt):
    new_var = self.rename_var(stmt.var)

    merge = self.transform_merge_before_loop(stmt.merge)
    new_start = self.transform_expr(stmt.start)
    new_stop = self.transform_expr(stmt.stop)
    new_step = self.transform_expr(stmt.step)
    new_body = self.transform_block(stmt.body)
    merge = self.transform_merge_after_loop(merge)
    return ForLoop(new_var, new_start, new_stop, new_step, new_body, merge)




from ..ndtypes import type_conv
from ..syntax import TupleProj, get_type, ActualArgs


from helpers import unpack_closure
from ..syntax.formal_args import FormalArgs, UnexpectedKeyword

def linearize_arg_types(fn, args):

  """
  Given a function object which might be one of:
    (1) a closure type
    (2) the name of an untyped function
    (3) an untyped fn object
  and some argument types which might
    (1) a list
    (2) a tuple
    (3) an ActualArgs object
  linearize the argument types with respect to
  the untyped function's argument order and return
  both the untyped function and list of arguments
  """

  untyped_fundef, closure_args = unpack_closure(fn)
  if isinstance(args, (list, tuple)):
    args = ActualArgs(args)
    
  if not isinstance(untyped_fundef.args, FormalArgs):
    assert False, "Expected function %s to have FormalArgs" % untyped_fundef.name 
    #formals = FormalArgs()
    #for arg in untyped_fundef.args:
    #  formals.add_positional(arg)
  else:
    formals = untyped_fundef.args

  if len(closure_args) > 0:
    args = args.prepend_positional(closure_args)

  def keyword_fn(_, v):
    return type_conv.typeof(v)

  try: 
    linear_args, extra = formals.linearize_values(args, keyword_fn = keyword_fn)
  except UnexpectedKeyword as e:
    e.fn_name = untyped_fundef.name  
    raise 
      
  return untyped_fundef, tuple(linear_args + extra)

def tuple_elts(tup):
  return [TupleProj(tup, i, t)
          for (i,t) in enumerate(tup.type.elt_types)]

def flatten_actual_args(args):
  if isinstance(args, (list,tuple)):
    return args
  assert isinstance(args, ActualArgs), \
      "Unexpected args: %s" % (args,)
  assert len(args.keywords) == 0
  result = list(args.positional)
  if args.starargs:
    result.extend(tuple_elts(args.starargs))
  return result

def linearize_actual_args(fn, args):
    untyped_fn, closure_args = unpack_closure(fn)
    if isinstance(args, (list, tuple)):
      args = ActualArgs(args)
    args = args.prepend_positional(closure_args)

    arg_types = args.transform(get_type)

    # Drop arguments that are assigned defaults,
    # since we're assuming those are set in the body
    # of the function
    if isinstance(untyped_fn.args, FormalArgs):
      try:
        combined_args = untyped_fn.args.linearize_without_defaults(args, tuple_elts)
      except UnexpectedKeyword as e:
        e.fn_name = fn.name 
        raise 
          
    else:
      combined_args = list(args)
    
    return untyped_fn, combined_args, arg_types


#
# Code taken from Numba's documentation at http://numba.pydata.org/numba-doc/0.11/arrays.html
#

import numpy as np 

mu = 0.1
Lx, Ly = 101, 101
N = 1000
import parakeet
import parakeet.c_backend
#parakeet.c_backend.config.print_module_source = True


def diffuse_loops(iter_num):
    u = np.zeros((Lx, Ly), dtype=np.float64)
    temp_u = np.zeros_like(u)
    temp_u[Lx / 2, Ly / 2] = 1000.0

    for n in range(iter_num):
        for i in range(1, Lx - 1):
            for j in range(1, Ly - 1):
                u[i, j] = mu * (temp_u[i + 1, j] + temp_u[i - 1, j] +
                                temp_u[i, j + 1] + temp_u[i, j - 1] -
                                4 * temp_u[i, j])

        temp = u
        u = temp_u
        temp_u = temp

    return u

def diffuse_array_expressions(iter_num):
    u = np.zeros((Lx, Ly), dtype=np.float64)
    temp_u = np.zeros_like(u)
    temp_u[Lx / 2, Ly / 2] = 1000.0

    for i in range(iter_num):
        u[1:-1, 1:-1] = mu * (temp_u[2:, 1:-1] + temp_u[:-2, 1:-1] +
                              temp_u[1:-1, 2:] + temp_u[1:-1, :-2] -
                              4 * temp_u[1:-1, 1:-1])

        temp = u
        u = temp_u
        temp_u = temp
    return u


from compare_perf import compare_perf 

compare_perf(diffuse_loops, [N], numba=True)
compare_perf( diffuse_array_expressions, [N], numba =True)




# Rosenbrock function derivative
# 
# Copied from https://github.com/numfocus/python-benchmarks/blob/master/rosen_der/rosen_der_python.py
# Original authors: Travis Oliphant (NumPy version) & Serge Guelton (loops version)
#

import numpy as np 
from parakeet import jit 

def rosen_der_np(x):
  der = np.empty_like(x)
  der[1:-1] = (+ 200 * (x[1:-1] - x[:-2] ** 2)
               - 400 * (x[2:] - x[1:-1] ** 2) * x[1:-1]
               - 2 * (1 - x[1:-1]))
  der[0] = -400 * x[0] * (x[1] - x[0] ** 2) - 2 * (1 - x[0])
  der[-1] = 200 * (x[-1] - x[-2] ** 2)
  return der

def rosen_der_loops(x):
  n = x.shape[0]
  der = np.empty_like(x)

  for i in range(1, n - 1):
    der[i] = (+ 200 * (x[i] - x[i - 1] ** 2)
              - 400 * (x[i + 1]
              - x[i] ** 2) * x[i]
              - 2 * (1 - x[i]))
  der[0] = -400 * x[0] * (x[1] - x[0] ** 2) - 2 * (1 - x[0])
  der[-1] = 200 * (x[-1] - x[-2] ** 2)
  return der

if __name__ == '__main__':
  N = 10**5
  x = np.arange(N) / float(N)
  jit(rosen_der_np)(x) 
  from compare_perf import compare_perf
  # numba still crashes on negative indexing
  compare_perf(rosen_der_np, [x.copy()], numba=False)
  compare_perf(rosen_der_loops, [x.copy()], numba=False)



import numpy as np
import parakeet
from parakeet import testing_helpers

 

@parakeet.jit
def range1(n):
  return range(n)

def test_range1():
  n = 10
  numpy_result = np.arange(n)
  parakeet_result = range1(n)
  testing_helpers.assert_eq_arrays(numpy_result, parakeet_result, "range1")
  
@parakeet.jit
def range2(start, stop):
  return range(start, stop)

def test_range2():
  shape = (100,111)
  numpy_result = np.arange(*shape)
  parakeet_result = range2(*shape)
  testing_helpers.assert_eq_arrays(numpy_result, parakeet_result, "range2")
  
@parakeet.jit
def range3(start, stop, step):
  return range(start, stop, step)

def test_range3():
  shape = (10,20,2)
  numpy_result = np.arange(*shape)
  parakeet_result = range3(10,20,2)
  testing_helpers.assert_eq_arrays(numpy_result, parakeet_result, "range3")


"""
def test_big_step():
  shape = (10,20,100)
  numpy_result =  np.arange(*shape)
  parakeet_result = range3(*shape)
  
  assert_eq_arrays(numpy_result, parakeet_result,"range_big_step")
"""

if __name__ == '__main__':
  testing_helpers.run_local_tests()




from ..analysis import collect_var_names
from ..builder import build_fn
from ..ndtypes import ArrayT 
from ..syntax import Var, ArrayExpr, Adverb, Range, Const, IndexMap 
from ..syntax.helpers import is_zero, is_one, is_false, is_true, is_none
from transform import Transform

class SimplifyArrayOperators(Transform):
  """
  Algebraic rewrite rules involving first-order array operations
  such as Range, ConstArray, etc.. 
  
  We do this conservatively by collecting all the array bindings
  and wiping them from the dictionary if there's even any chance they 
  might get modified
  """
  
  def pre_apply(self, fn):
    self.bindings = []
    return fn 
  
  def push(self):
    self.bindings.append({})
    
  def pop(self):
    self.bindings.pop()
    
  def mark_dirty(self, name):
    if isinstance(self.type_env[name], ArrayT):
      for d in self.bindings:
        if name in d:
          del d[name]
  def bind(self, name, v):
    self.bindings[-1][name] = v 
    
  def lookup(self, name):
    return self.bindings[-1].get(name)
  
  def transform_Attribute(self, expr):
    return expr
  
  def transform_Var(self, expr):
    """
    If we're using an array variable *anywhere* assume it might get modified
    """
    self.mark_dirty(expr.name)
    return expr  
  
  def transform_Map(self, expr):
    """
    Map(f, Range(start,stop)) = IndexMap(f', (stop - start))
    Map(f, NDIndex(shape)) = IndexMap(f, shape)
    """
     
    if len(expr.args) != 1:
      return expr 
    arg = expr.args[0]
    if arg.__class__ is not Var:
      return expr 
    name = arg.name
    prev = self.lookup(name)
    if prev is None:
      return expr 
    if prev.__class__ is Range:
      step = prev.step   
      if step.__class__ is not Const:
        return expr 
      if step.value is not None and step.value != 1:
        return expr 
      diff = self.sub(prev.stop, prev.start, "niters")
      return IndexMap(fn = expr.fn, shape = diff, start_index = prev.start, type = expr.type) 
    return expr 
      
  def transform_Assign(self, stmt):
    
    if isinstance(stmt.lhs, Var):
      if isinstance(stmt.rhs, (ArrayExpr, Adverb)):
        self.bind(stmt.lhs.name, stmt.rhs)
      stmt.rhs = self.transform_expr(stmt.rhs)
    else:
      for name in collect_var_names(stmt.lhs):
        self.mark_dirty(name)
      for name in collect_var_names(stmt.rhs):
        self.mark_dirty(name)
    return stmt 
  
  def transform_block(self, stmts):
    self.push()
    stmts = Transform.transform_block(self, stmts)
    self.pop()
    return stmts 


import numpy as np 

from .. import names
from ..builder import build_fn
# from ..c_backend import FnCompiler
from ..ndtypes import TupleT,  Int32, FnT, ScalarT, SliceT, NoneT, ArrayT, ClosureT
from ..c_backend import PyModuleCompiler
from ..openmp_backend import MulticoreCompiler
from ..syntax import PrintString, SourceExpr
from ..syntax.helpers import get_types, get_fn, get_closure_args, const_int, zero_i32, one_i32

import config 
import device_info 
from cuda_syntax import threadIdx, blockIdx, blockDim 


class CudaCompiler(MulticoreCompiler):
  
  def __init__(self, *args, **kwargs):
    # keep track of which kernels we've already compiled 
    # and map the name of the nested function to its kernel name
    self._kernel_cache = {}
    if 'gpu_depth' in kwargs:
      self.gpu_depth = kwargs['gpu_depth'] 
      del kwargs['gpu_depth']
    else:
      self.gpu_depth = 0
    
    self.device = device_info.best_cuda_device()
    assert self.device, "No GPU found"
    
    self.use_constant_memory_for_args = True
    MulticoreCompiler.__init__(self, 
                               compiler_cmd = ['nvcc', '-arch=%s' % config.arch],
                               extra_link_flags = ['-lcudart'], 
                               src_extension = '.cu',  
                               compiler_flag_prefix = '-Xcompiler',
                               linker_flag_prefix = '-Xlinker', 
                               *args, **kwargs)
  
  @property 
  def cache_key(self):
    return self.__class__, self.depth > 0, max(self.gpu_depth, 2) 
  
  def enter_module_body(self):
    self.append('cudaSetDevice(%d);' % device_info.device_id(self.device))
  
  def build_kernel(self, clos, bounds):
    n_indices = len(bounds)
    fn = get_fn(clos)
    outer_closure_exprs = get_closure_args(clos)
    closure_arg_types = get_types(outer_closure_exprs)
    host_closure_args = self.visit_expr_list(outer_closure_exprs)
    
    self.comment("Copying data from closure arguments to the GPU")
    
    
    read_only = [False] * len(host_closure_args)
    write_only = [False] * len(host_closure_args)
    
    gpu_closure_args = self.args_to_gpu(host_closure_args, closure_arg_types, write_only)
    input_types = fn.input_types 

    kernel_name = names.fresh("kernel_" + fn.name)
    
    if isinstance(input_types[-1], TupleT):
      index_types = input_types[-1].elt_types   
      index_as_tuple = True
      # if function takes a tuple of 
      closure_arg_types = input_types[:-1]
    else:
      index_types = input_types[-n_indices:]
      index_as_tuple = False
      closure_arg_types = input_types[:-n_indices]
      
    n_closure_args = len(closure_arg_types)
    assert len(outer_closure_exprs) == n_closure_args, \
      "Mismatch between closure formal args %s and given %s" % (", ".join(closure_arg_types),
                                                                ", ".join(outer_closure_exprs))
    bound_types = (Int32,) * n_indices
    
    if self.use_constant_memory_for_args:
      outer_input_types = tuple(bound_types)
    else:
      outer_input_types = tuple(closure_arg_types) + bound_types
      
    
        
    
    parakeet_kernel, builder, input_vars  = build_fn(outer_input_types, name = kernel_name)
    
    if self.use_constant_memory_for_args:
      inner_closure_vars = []
      for i, t in enumerate(closure_arg_types):
        raw_name = fn.arg_names[i]
        self.comment("Moving GPU arg #%d %s : %s to constant memory" % (i, raw_name, t))
        const_name = self.fresh_name("gpu_arg_"  + raw_name)
        typename = self.to_ctype(t)
        self.add_decl("__constant__ %s %s" % (typename, const_name))
        inner_closure_vars.append(SourceExpr(const_name, type=t))
        gpu_arg = gpu_closure_args[i]
        # in case this is a constant, should assign to a variable
        first_char = gpu_arg[0]
        if not first_char.isalpha():
          gpu_arg = self.fresh_var(typename, "src_" + raw_name, gpu_arg)
        self.append("cudaMemcpyToSymbolAsync(%s, &%s, sizeof(%s));" % (const_name, gpu_arg, typename))
      inner_closure_vars = tuple(inner_closure_vars)
      # need to do a cudaMemcpyToSymbol for each gpu arg   
      bounds_vars = input_vars
  
    else:
      # TODO: use these to compute indices when n_indices > 3 or 
      # number of threads per block > 1  
      inner_closure_vars = input_vars[:n_closure_args]
      bounds_vars = input_vars[n_closure_args:(n_closure_args + n_indices)]


    THREADS_PER_DIM = config.threads_per_block_dim 

    
    if n_indices == 1:
      elts_per_block = builder.div(bounds_vars[0],  
                                   const_int(THREADS_PER_DIM**2, Int32), 
                                   "elts_per_block")
      elts_per_block = builder.add(elts_per_block, one_i32, "elts_per_block")
      base_idx = builder.mul(elts_per_block, blockIdx.x, "base_x")
      start_idx = builder.add(base_idx, threadIdx.x, "start_x")  
      
      start_indices = [start_idx]
    
      stop_x = builder.mul(builder.add(one_i32, blockIdx.x, "next_base_x"), 
                           elts_per_block, "stop_x")
      stop_x = builder.min(stop_x, bounds_vars[0], "stop_x") 
      stop_indices = [stop_x]
      step_sizes = [const_int(THREADS_PER_DIM**2, Int32)]
       
    else:
      elts_per_block = builder.div(bounds_vars[0],   
                                   const_int(THREADS_PER_DIM, Int32),
                                    "elts_per_block")
      elts_per_block = builder.add(elts_per_block, one_i32, "elts_per_block")
      base_x = builder.mul(elts_per_block, blockIdx.x, "base_x")
      start_x = builder.add(base_x, threadIdx.x, "start_x")  
      start_y = builder.add(builder.mul(elts_per_block, blockIdx.y, "base_y"),
                                threadIdx.y, "start_y")
      
      start_indices = [start_x, start_y]
      
      stop_x = builder.mul(builder.add(one_i32, blockIdx.x, "next_base_x"), 
                           elts_per_block, "stop_x")
      stop_x = builder.min(stop_x, bounds_vars[0], "stop_x") 
      stop_indices = [stop_x]
      stop_y = builder.mul(builder.add(one_i32, blockIdx.y, "next_base_y"),
                           elts_per_block, "stop_y")
      stop_y = builder.min(stop_y, bounds_vars[1], "stop_y")
       
      stop_indices = [stop_x, stop_y]
      
      step_sizes = [const_int(THREADS_PER_DIM, Int32), const_int(THREADS_PER_DIM, Int32)]
      for i in xrange(2, n_indices):
        start_indices.append(zero_i32)
        step_sizes.append(one_i32)
        stop_indices.append(bounds_vars[i])
    
    
    
    def loop_body(index_vars):
      if not isinstance(index_vars, (list,tuple)):
        pass
      indices = [builder.cast(idx, t) for idx, t in zip(index_vars,index_types)]
      if index_as_tuple:
        index_args = (builder.tuple(indices),)
      else:
        index_args = indices
      inner_args = tuple(inner_closure_vars) + tuple(index_args)
      builder.call(fn, inner_args)
      
    builder.nested_loops(stop_indices, loop_body, start_indices, step_sizes, index_vars_as_list = True)
    
    self.enter_kernel()
    c_kernel_name = self.get_fn_name(parakeet_kernel, 
                                     attributes = ["__global__"], 
                                     inline = False)
    self.exit_kernel()
    
    # set cache preference of the kernel we just built 
    self.append("cudaFuncSetCacheConfig(%s, cudaFuncCachePreferL1);" % c_kernel_name)
    # self._kernel_cache[key] = c_kernel_name
    return c_kernel_name, gpu_closure_args, host_closure_args, closure_arg_types
  
  def launch_kernel(self, bounds, params, kernel_name):
    self.synchronize("Done copying arguments to GPU, prepare for kernel launch")
    
    n_bounds = len(bounds)
    sm_count = device_info.num_multiprocessors(self.device)
    n_blocks = sm_count * config.blocks_per_sm 
    
    THREADS_PER_DIM = config.threads_per_block_dim
    
    if n_bounds == 1:
      grid_dims = [n_blocks, 1, 1]
      block_dims = [THREADS_PER_DIM**2, 1, 1]
    else:
      blocks_per_axis = int(np.ceil(np.sqrt(n_blocks)))
      grid_dims = [blocks_per_axis, blocks_per_axis, 1]
      block_dims = [THREADS_PER_DIM,THREADS_PER_DIM,1]

    grid_dims_str = "dim3(%s)" % ", ".join( str(d) for d in grid_dims)
    block_dims_str = "dim3(%s)" % ", ".join( str(d) for d in block_dims)
    
    self.comment("kernel launch")
    
    kernel_args_str = ", ".join(params)
    self.append("%s<<<%s, %s>>>(%s);" % (kernel_name, grid_dims_str, block_dims_str, kernel_args_str))
  
    self.comment("After launching kernel, synchronize to make sure the computation is done")
    self.synchronize("Finished kernel launch")
    
    
  
  def visit_ParFor(self, stmt):
    bounds = self.tuple_to_var_list(stmt.bounds)

    n_indices = len(bounds)
    if n_indices > 5 or not self.in_host():
      return MulticoreCompiler.visit_ParFor(self, stmt)

    
    kernel_name, gpu_closure_args, host_closure_args, closure_arg_types  = \
      self.build_kernel(stmt.fn, bounds)
    
    if self.use_constant_memory_for_args:
      params = bounds 
    else:
      params = tuple(gpu_closure_args) + tuple(bounds)
      
    self.launch_kernel(bounds, params, kernel_name)
    
    self.comment("copy arguments back from the GPU to the host")
    self.list_to_host(host_closure_args, gpu_closure_args, closure_arg_types)
    return "/* done with ParFor */"
  
  def in_host(self):
    return self.gpu_depth == 0
  
  def in_block(self):
    return self.gpu_depth == 1
  
  def in_gpu(self):
    return self.gpu_depth > 0
  
  def get_fn_name(self, fn_expr, attributes = [], inline = True):
    if self.in_gpu() and not attributes:
      attributes = ["__device__"] 
    kwargs = {'depth':self.depth, 'gpu_depth':self.gpu_depth}
    return PyModuleCompiler.get_fn_name(self, fn_expr, 
                                        compiler_kwargs = kwargs,
                                        attributes = attributes, 
                                        inline = inline)
    

  
  def enter_kernel(self):
    """
    Keep a stack of adverb contexts so we know when we're global vs. block vs. thread
    """
    self.depth += 1
    self.gpu_depth += 1  
  
  def exit_kernel(self):
    self.depth -= 1
    self.gpu_depth -= 1 
  
  
  def in_thread(self):
    return self.depth > 1
  
  def pass_by_value(self, t):
    if isinstance(t, (ScalarT, NoneT, SliceT, FnT)):
      return True 
    elif isinstance(t, TupleT):
      return all(self.pass_by_value(elt_t) for elt_t in t.elt_types)
    elif isinstance(t, ClosureT):
      return all(self.pass_by_value(elt_t) for elt_t in t.arg_types)
    return False 
  
  
  def check_gpu_error(self, context = None, error_code_var = None):
    if error_code_var is None:
      error_code_var = self.fresh_name("cuda_err")
      self.append("cudaError %s = cudaGetLastError();" % error_code_var)
    if context is None:
      context = "\"Generated CUDA source at \" __FILE__ \" : \" __LINE__"
    self.append("""
      if ( cudaSuccess != %s ) {
        printf( "Error after %s: %%s\\n",  cudaGetErrorString(%s) );
      }
    """ % (error_code_var, context, error_code_var))
    
    if self.module_entry:
      self.append("""
        /* TODO: use setjmp/longjmp or C++ exceptions to propagate exceptions 
           outside of the entry function 
        */ 
        if (cudaSuccess != %s) {
          PyErr_Format(PyExc_SystemError, 
                      "Parakeet GPU Error after %%s: %%s", 
                      "%s", 
                      cudaGetErrorString(%s)); 
          return NULL; 
        }
      """ % (error_code_var, context, error_code_var))
          
  
  def synchronize(self, context = None):
    error_code_var = self.fresh_name("cuda_err")
    self.append("cudaError %s = cudaDeviceSynchronize();" % error_code_var)
    self.check_gpu_error(context, error_code_var)
    
  
  def _to_gpu(self, c_expr, t, gpu_array_dict, memcpy = True):
    if self.pass_by_value(t):
      return c_expr
    elif isinstance(t, ArrayT):
      
      ptr_t = "%s*" % self.to_ctype(t.elt_type)
      bytes_per_elt = t.elt_type.dtype.itemsize
      
      dst = self.fresh_var(ptr_t, "gpu_ptr")
      nelts = "%s.size" % c_expr
      nbytes = self.fresh_var("int64_t", "nbytes", "%s * %d" % (nelts, bytes_per_elt))
      
      if memcpy:
        src = "%s.data.raw_ptr" % c_expr   
        memcpy_stmt = "cudaMemcpyAsync(%s, %s, %s, cudaMemcpyHostToDevice);" % (dst, src, nbytes)
      else:
        memcpy_stmt = "/* no memcpy for this %s : %s */" % (c_expr, t)
      
      # allocate the destination pointer on the GPU
      malloc = "cudaMalloc( (void**) &%s, %s);" % (dst, nbytes)
      alloc_and_copy_block = "%s\n%s" % (malloc, memcpy_stmt)

      """
      TODO: do this to cut down on allocations but also
      need to track which arrays have been allocated so we 
      can delete them later 
      
      if t in gpu_array_dict:
        # if we've allocated some other args of the same type, 
        # try reusing their pointers before allocating new memory 
        for i, (prev_host, prev_gpu) in enumerate(gpu_array_dict[t]):
          cond = "%s.data.raw_ptr == %s.data.raw_ptr && %s.size == %s.size" % \
            (c_expr, prev_host, c_expr, prev_host)
          self.append("%sif (%s) {%s = %s.data.raw_ptr;}" % \
                        ("else " if i > 0 else "", cond, dst, prev_gpu))
        self.append("else {%s}" % alloc_and_copy_block)
      else:
      """
      self.append(alloc_and_copy_block)
      
      self.check_gpu_error("cudaMalloc for %s : %s" % (c_expr, t))
      
      # make an identical array descriptor but change its data pointer to the GPU location
      gpu_descriptor = self.fresh_var(self.to_ctype(t), "gpu_array", c_expr)
      self.append("%s.data.raw_ptr = %s;" % (gpu_descriptor, dst))    
      gpu_array_dict.setdefault(t, []).append( (c_expr, gpu_descriptor) )
      return gpu_descriptor
    
    elif isinstance(t, (ClosureT, TupleT)):
      # copy contents of the host tuple into another struct
      gpu_tuple = self.fresh_var(self.to_ctype(t), "gpu_tuple", c_expr)
      for i, elt_t in enumerate(t.elt_types):
        host_elt = "%s.elt%d" % (c_expr, i)
        gpu_elt = self._to_gpu(host_elt, elt_t, gpu_array_dict)
        self.append("%s.elt%d = %s;" % (c_expr, i, gpu_elt))
      return gpu_tuple 
    else:
      assert False, "Unsupported type in CUDA backend %s" % t 
  
  def args_to_gpu(self, host_values, types, write_only):
    # keep track of arrays we've already moved to the GPU, 
    # indexed by their type  
    gpu_array_dict = {}
    
    gpu_args = []
    for i,t in enumerate(types):
      gpu_arg = self._to_gpu(host_values[i], t, gpu_array_dict, memcpy = not write_only[i])
      gpu_args.append(gpu_arg)
    return gpu_args 
  
  
  def to_host(self, host_value, gpu_value, t):
    if self.pass_by_value(t):
      return
    elif isinstance(t, ArrayT):
      dst = "%s.data.raw_ptr" % host_value 
      src = "%s.data.raw_ptr"  % gpu_value 
      nelts = "%s.size" % gpu_value
      nbytes = "%s * %d" % (nelts, t.elt_type.dtype.itemsize)  
      self.append("cudaMemcpy(%s, %s, %s, cudaMemcpyDeviceToHost);" % (dst, src, nbytes) )
      self.append("cudaFree(%s);" % src) 

    elif isinstance(t, (ClosureT, TupleT)):
      for i, elt_t in enumerate(t.elt_types):
        host_elt = "%s.elt%d" % (host_value, i)
        gpu_elt = "%s.elt%d" % (gpu_value, i)
        self.to_host(host_elt, gpu_elt, elt_t)
    else:
      assert False, "Unsupported type in CUDA backend %s" % t 
      
  def list_to_host(self, host_values, gpu_values, types):
    for i, t in enumerate(types):
      host_value = host_values[i]
      gpu_value = gpu_values[i]
      self.to_host(host_value, gpu_value, t)
    
  def visit_Alloc(self, expr):
    assert self.in_host(), "Can't dynamically allocate memory in GPU code"
    return MulticoreCompiler.visit_Alloc(self, expr)
  
  def visit_AllocArray(self, expr):
    assert self.in_host(), "Can't dynamically allocate memory in GPU code"
    return MulticoreCompiler.visit_AllocArray(self, expr)

  """
  def visit_IndexReduce(self, expr):
    fn =  self.get_fn(expr.fn, qualifier = "device")
    combine = self.get_fn(expr.combine, qualifier = "device")
    if self.in_host():
      #
      # weave two device functions together into a  reduce kernel
      #
      pass 
 
  
  def visit_IndexScan(self, expr):
    fn =  self.get_fn(expr.fn, qualifier = "device")
    combine = self.get_fn(expr.combine, qualifier = "device")
    emit = self.get_fn(expr.emit, qualifier = "device")
    if self.in_host():
      pass 
  """




from collect_vars import (collect_binding_names, 
                          collect_bindings, 
                          collect_var_names, 
                          collect_var_names_from_exprs, 
                          collect_var_names_list)
from contains import (contains_adverbs, contains_calls, contains_functions, 
                      contains_loops, contains_slices, contains_structs, 
                      contains_array_operators)
 
from escape_analysis import may_alias, may_escape, escape_analysis 

from find_local_arrays import FindLocalArrays
from index_elim_analysis import IndexElimAnalysis
from inline_allowed import can_inline
from mutability_analysis import find_mutable_types, TypeBasedMutabilityAnalysis
from offset_analysis import OffsetAnalysis 
from syntax_visitor import SyntaxVisitor 
from use_analysis import find_live_vars, use_count
from usedef import StmtPath, UseDefAnalysis
from value_range_analysis import ValueRangeAnalyis
from verify import verify 






from parakeet import jit, config, c_backend 
 


def covariance(x,y):
  return ((x-x.mean()) * (y-y.mean())).mean()

def fit_simple_regression(x,y):
  slope = covariance(x,y) / covariance(x,x)
  offset = y.mean() - slope * x.mean() 
  return slope, offset

import numpy as np 

N = 10**7
x = np.random.randn(N).astype('float64')
slope = 903.29
offset = 102.1
y = slope * x + offset



from compare_perf import compare_perf 
compare_perf(fit_simple_regression, (x,y))



from .. import names 

from ..builder import Builder
from ..ndtypes import (ScalarT, NoneT, NoneType, ArrayT, SliceT, TupleT, make_tuple_type, 
                       Int64, PtrT, ptr_type, ClosureT, FnT, StructT, 
                       TypeValueT)
from ..syntax import (Var, Tuple, 
                      Index, TypedFn, Return, Stmt, Assign, Alloc,  
                      ParFor, PrimCall, If, While, ForLoop, Call, Expr, 
                      IndexReduce, ExprStmt) 
from ..syntax.helpers import none, const_int

from transform import Transform

def concat(seqs):
  result = []
  for seq in seqs:
    for elt in seq:
      result.append(elt)
  return tuple(result)

def concat_args(*seqs):
  return concat(seqs)

def concat_map(f, seq):
  return concat(f(elt) for elt in seq)

def flatten_type(t):
  if isinstance(t, (ScalarT, PtrT)):
    return (t,)
  elif isinstance(t, TupleT):
    return concat(flatten_type(elt_t) for elt_t in t.elt_types)
  elif isinstance(t, (NoneT, FnT, TypeValueT)):
    return ()
  elif isinstance(t, ClosureT):
    return concat(flatten_type(elt_t) for elt_t in t.arg_types)
  elif isinstance(t, ArrayT):
    return concat_args(
      flatten_type(t.ptr_t), 
      flatten_type(t.shape_t),
      flatten_type(t.strides_t),          
      (Int64, Int64), # offset and size
      
    )
  elif isinstance(t, SliceT):
    return flatten_types( (t.start_type, t.stop_type, t.step_type) )
  else:
    assert False, "Unsupported type %s" % (t,)

def flatten_types(ts):
  return concat([flatten_type(t) for t in ts])
  
def field_pos_range(t, field, _cache = {}):
  key = (t, field)
  if key in _cache:
    return _cache[key]
  assert isinstance(t, StructT), "Expected struct got %s.%s" % (t, field)
  offset = 0
  for i, (field_name, field_t) in enumerate(t._fields_):
    n = len(flatten_type(field_t))
    if field_name == field or (isinstance(field, (int, long)) and i == field):
      result = (offset, offset+n)
      _cache[key] = result
      return result
    offset += n 
  assert False, "Field %s not found on type %s" % (field, t)

def get_field_elts(t, values, field):
  start, stop = field_pos_range(t, field)
  assert stop <= len(values), \
    "Insufficient number of flattened fields %s for %s.%s" % (values, t, field)
  return values[start:stop]

def single_type(ts):
  """
  Turn a sequence of types into a single type object
  """
  if len(ts) == 0:
    return NoneType
  elif len(ts) == 1:
    return ts[0]
  else:
    return make_tuple_type(ts)
  
def single_value(values):
  if len(values) == 0:
    return none 
  elif len(values) == 1:
    return values[0]
  else:
    t = make_tuple_type(tuple(v.type for v in values))
    return Tuple(values, type = t)
  
def mk_vars(names, types):
  """
  Combine a list of names and a list of types into a single list of vars
  """
  return [Var(name = name, type = t) for name, t in zip(names, types)]


class BuildFlatFn(Builder):
  def __init__(self, old_fn):
    Builder.__init__(self)
    self.old_fn = old_fn 
    base_name = names.original(old_fn.name)
    flat_fn_name = names.fresh("flat_" + base_name)

    self.type_env = {}
    for (name, t) in old_fn.type_env.iteritems():

      old_var = Var(name = name, type = t)
      for new_var in self.flatten_lhs_var(old_var):
  
        self.type_env[new_var.name] = new_var.type
    old_input_vars = mk_vars(old_fn.arg_names, old_fn.input_types)
        
    # a var of struct type from the old fn 
    # maps to multiple variables in the new
    # function 
    self.var_expansions = {}
    
    new_input_vars = []
    for var in old_input_vars:
      flat_vars = self.flatten_lhs_var(var)
      self.var_expansions[var.name]  = flat_vars 
      new_input_vars.extend(flat_vars)
    
    new_input_names = tuple(var.name for var in new_input_vars)
    new_input_types = tuple(var.type for var in new_input_vars)
    new_return_type =  single_type(flatten_type(old_fn.return_type))
    
    self.flat_fn = \
      TypedFn(name = flat_fn_name, 
        arg_names = new_input_names, 
        body = [], 
        type_env = self.type_env,
        input_types = new_input_types, 
        return_type = new_return_type, 
        created_by = self) 
      
    
  def run(self):
    self.flat_fn.body = self.flatten_block(self.old_fn.body)
    return self.flat_fn
  
  
  #######################
  #
  #     Helpers
  #
  #######################
  
  def flatten_block(self, stmts):
    self.blocks.push()
    for stmt in stmts:
      result = self.flatten_stmt(stmt)
      if result is None:
        continue
      if not isinstance(result, (list,tuple)):
        result = [result]
      for new_stmt in result:
        assert isinstance(new_stmt, Stmt), "Expected statement, got %s" % (new_stmt,)
        self.blocks.append(new_stmt)
    return self.blocks.pop()
  
  
  
  #######################
  #
  #     Statements
  #
  #######################
  
  
  #def bind_name(self):
  #def bind_var(self, lhs, rhs):
  
  def flatten_Assign(self, stmt):
    c = stmt.lhs.__class__
    rhs = self.flatten_expr(stmt.rhs)

    if c is Var:
      
      lhs_vars = self.flatten_lhs_var(stmt.lhs)
      self.var_expansions[stmt.lhs.name] = lhs_vars 
      if isinstance(rhs, Expr):
        if len(lhs_vars) == 1:
          return [Assign(lhs_vars[0], rhs)]
        else:
          # the IR doesn't allow for multiple assignment 
          # so we fake it with tuple literals
          return [Assign(self.tuple(lhs_vars), rhs)]
      assert isinstance(rhs, (list,tuple))
      assert len(lhs_vars) == len(rhs), \
        "Mismatch between LHS %s and RHS %s : %s => %s in stmt %s" % \
        (lhs_vars, stmt.rhs, stmt.rhs.type, rhs, stmt)
      result = []
      for var, value in zip(lhs_vars, rhs):
        result.append(Assign(var, value))
      return result 

    elif c is Index:
      array_t = stmt.lhs.value.type 
      if isinstance(array_t, PtrT):
        return stmt  
      indices = self.flatten_expr(stmt.lhs.index)
      values = self.flatten_expr(stmt.lhs.value)
      data = get_field_elts(array_t, values, 'data')[0]
      shape = get_field_elts(array_t, values, 'shape')
      strides = get_field_elts(array_t, values, 'strides')
      offset = get_field_elts(array_t, values, 'offset')[0]
      n_dims = len(strides)
      n_indices = len(indices)
      assert n_dims == n_indices, \
        "Expected %d indices but only got %d in %s" % (n_dims, n_indices, stmt)
      for idx, stride in zip(indices, strides):
        offset = self.add(offset, self.mul(idx, stride))

      stmt.lhs = self.index(data, offset, temp=False)
      stmt.rhs = rhs[0]
      return [stmt]
    else:
      assert False, "LHS not supported in flattening: %s" % stmt 
  
  def enter_branch(self, phi_nodes):
    for (k, (left, _)) in phi_nodes.iteritems():
      self.var_expansions[k] = self.flatten_lhs_name(k, left.type)
  
  def flatten_merge(self, phi_nodes):
    result = {}
    for (k, (left, right)) in phi_nodes.iteritems():
      t = left.type
      assert right.type == t 
      if isinstance(t, (ScalarT, PtrT)):
        result[k] = (self.flatten_scalar_expr(left), self.flatten_scalar_expr(right))
      elif isinstance(t, (FnT, NoneT)):
        continue 
      else:
        fields = self.var_expansions[k]
        flat_left = self.flatten_expr(left)
        flat_right = self.flatten_expr(right)
        assert len(fields) == len(flat_left)
        assert len(fields) == len(flat_right)
        for i, var in enumerate(fields):
          result[var.name] = (flat_left[i], flat_right[i])
    return result 
   
  def flatten_ForLoop(self, stmt):
    self.enter_branch(stmt.merge)
    var = self.flatten_scalar_lhs_var(stmt.var)
    start = self.flatten_scalar_expr(stmt.start)
    stop = self.flatten_scalar_expr(stmt.stop)
    step = self.flatten_scalar_expr(stmt.step)
    body = self.flatten_block(stmt.body)
    merge = self.flatten_merge(stmt.merge)
    return ForLoop(var, start, stop, step, body, merge)
  
  def flatten_While(self, stmt):
    self.enter_branch(stmt.merge)
    cond = self.flatten_scalar_expr(stmt.cond)
    body = self.flatten_block(stmt.body)
    merge = self.flatten_merge(stmt.merge)
    return While(cond, body, merge)
     
  
  def flatten_If(self, stmt):
    self.enter_branch(stmt.merge)
    cond = self.flatten_scalar_expr(stmt.cond)
    true = self.flatten_block(stmt.true)
    false = self.flatten_block(stmt.false)
    merge = self.flatten_merge(stmt.merge)
    assert merge is not None
    return If(cond, true, false, merge = merge)
  
  def flatten_ExprStmt(self, stmt):
    return ExprStmt(value = self.flatten_expr(stmt.value))
   
  def flatten_ParFor(self, stmt):
    new_fn, closure_elts = self.flatten_fn(stmt.fn)
    closure = self.closure(new_fn, closure_elts)
    bounds = single_value(self.flatten_expr(stmt.bounds))
    return ParFor(fn = closure, bounds = bounds)
    
  def flatten_Return(self, stmt):
    return Return(single_value(self.flatten_expr(stmt.value)))
  
  def flatten_Comment(self, stmt):
    return stmt 
  
  def flatten_stmt(self, stmt):
    method_name = "flatten_%s" % stmt.__class__.__name__
    return getattr(self, method_name)(stmt)
  
  #######################
  #
  #     Expressions 
  #
  #######################
  
  def flatten_expr(self, expr):
    method_name = "flatten_%s" % expr.__class__.__name__
    return getattr(self, method_name)(expr)
  
  def flatten_expr_list(self, exprs):
    return concat_map(self.flatten_expr, exprs)
  
  def flatten_scalar_expr(self, expr):
    """
    Give me back a single expression instead of a list 
    """
    flat_exprs = self.flatten_expr(expr)
    assert len(flat_exprs) == 1
    return flat_exprs[0]
    
  def flatten_scalar_expr_list(self, exprs):
    assert isinstance(exprs, (list, tuple)), "Expected list, got %s" % (exprs,)
    return [self.flatten_scalar_expr(e) for e in  exprs]
  
  def flatten_Const(self, expr):
    if isinstance(expr.type, NoneT):
      return ()
    return (expr,)
  
  def flatten_fn(self, closure):
    fn = self.get_fn(closure)
    import pipeline
    fn = pipeline.indexify.apply(fn)
    flat_fn = build_flat_fn(fn)
    flat_closure_args = self.flatten_expr(closure)
    return flat_fn, flat_closure_args
  
  
  def flatten_Call(self, expr):
    flat_fn, flat_closure_args = self.flatten_fn(expr.fn)
    flat_args = self.flatten_expr_list(expr.args)
    args = tuple(flat_closure_args) + tuple(flat_args)
    return Call(flat_fn, args, type = flat_fn.return_type)
  
  def flatten_Cast(self, expr):
    return [expr]
  
  def flatten_UntypedFn(self, expr):
    return []
  
  def flatten_TypedFn(self, expr):
    return []
    
  def flatten_Var(self, expr):
    if isinstance(expr.type, (ScalarT, PtrT)):
      return (expr,)
    elif isinstance(expr.type, (FnT, NoneT)):
      return ()
    else:
      name = expr.name
      assert name in self.var_expansions, "No fields known for %s : %s" % (expr, expr.type) 
      return self.var_expansions[name]
    
  def flatten_Tuple(self, expr):
    return self.flatten_expr_list(expr.elts)
  
  def flatten_field(self, struct, field):
    elts = self.flatten_expr(struct)
    start, stop = field_pos_range(struct.type, field)
    return elts[start:stop]
   
  def flatten_TupleProj(self, expr):
    result = self.flatten_field(expr.tuple, expr.index)
    return result
    
  def flatten_Closure(self, expr):
    return self.flatten_expr_list(expr.args)
    
  def flatten_ClosureElt(self, expr):
    return self.flatten_field(expr.closure, expr.index)
  
  def flatten_Attribute(self, expr):
    return self.flatten_field(expr.value, expr.name)
  
  def flatten_Select(self, expr):
    cond = self.flatten_scalar_expr(expr.cond)
    true_values = self.flatten_expr(expr.true_value)
    false_values = self.flatten_expr(expr.false_value)
    assert len(true_values) == len(false_values)
    return [self.select(cond,t,f) for t,f in zip(true_values, false_values)]
    
  def flatten_Alloc(self, expr):
    count_exprs = self.flatten_expr(expr.count)
    assert len(count_exprs) == 1
    return Alloc(count = count_exprs[0], elt_type = expr.elt_type, type = expr.type)
    
  def flatten_Array(self, expr):
    assert False, "Array node should be an explicit allocation by now"
    # or, if we flatten structured elts, maybe we should handle it here?
  
    
  def flatten_Index(self, expr):

    t = expr.value.type 
    if isinstance(t, PtrT):
      return [expr]   
    assert isinstance(t, ArrayT), "Expected Index to take array, got %s" % (expr.type,)
    array_fields = self.flatten_expr(expr.value)
    data_fields = get_field_elts(t, array_fields, 'data')
    shape = get_field_elts(t, array_fields, 'shape')
    strides = get_field_elts(t, array_fields, 'strides')
    offset = get_field_elts(t, array_fields, 'offset')[0]
    
    index = expr.index 
    if isinstance(index.type, (NoneT, SliceT, ScalarT)):
      indices = [index]
    elif isinstance(index, Tuple):
      indices = index.elts 
    else:
      assert isinstance(index.type, TupleT), "Expected index to scalar, slice, or tuple"
      indices = self.tuple_elts(index)


      
    #indices = self.flatten_expr(expr.index)
   
    n_indices = len(indices)
    n_strides = len(strides)
    assert n_indices == n_strides, \
      "Not supported:  indices vs. dimensions: %d != %d in %s" % (n_indices, n_strides, expr)


    # fast-path for the common case when we're indexing
    # by all scalars to retrieve a scalar result
    #if syntax.helpers.all_scalars(indices):
    for i, idx in enumerate(indices):
      offset = self.add(offset, self.mul(idx, strides[i]))
    return [self.index(data_fields[0], offset)]
     
  
  def flatten_PrimCall(self, expr):
    args = self.flatten_scalar_expr_list(expr.args)
    return [PrimCall(prim = expr.prim, args = args, type = expr.type)]
  
  def flatten_Slice(self, expr):
    return self.flatten_expr_list([expr.start, expr.stop, expr.step])
  
  def flatten_Len(self, expr):
    assert False, "Not implemented" 
  
  def flatten_ConstArray(self, expr):
    assert False, "Not implemented" 
  
  def flatten_ConstArrayLike(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Range(self, expr):
    assert False, "Not implemented" 
  
  def strides_from_shape_elts(self, shape_elts):
    strides = [const_int(1)]
    for dim in reversed(shape_elts[1:]):
      strides = [self.mul(strides[0], dim)] + strides
    return strides
  
  def flatten_AllocArray(self, expr):
    nelts = const_int(1)
    shape_elts = self.flatten_expr(expr.shape)
    for dim in shape_elts:
      nelts = self.mul(nelts, dim)
    ptr = Alloc(elt_type = expr.elt_type, count = nelts, type = ptr_type(expr.elt_type))
    stride_elts = self.strides_from_shape_elts(shape_elts)
    return (ptr,) + tuple(shape_elts) + tuple(stride_elts) + (self.int(0), nelts)
  
  def flatten_ArrayView(self, expr):
    data = self.flatten_expr(expr.data)
    shape = self.flatten_expr(expr.shape)
    strides = self.flatten_expr(expr.strides)
    offset = self.flatten_expr(expr.offset)
    size = self.flatten_expr(expr.size)
    return data + shape + strides + offset + size
    
  
  def flatten_Ravel(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Reshape(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Shape(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Strides(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Transpose(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Where(self, expr):
    assert False, "Not implemented" 
     
   
    
  #######################
  #
  # Adverbs
  #
  #######################
  def flatten_Map(self, expr):
    assert False, "Unexpected Map encountered during flattening, should be IndexScan"
  
  def flatten_Reduce(self, expr):
    assert False, "Unexpected Reduce encountered during flattening, should be IndexScan"
  
  def flatten_Scan(self, expr):
    assert False, "Unexpected Scan encountered during flattening, should be IndexScan"
  
  def flatten_IndexMap(self, expr):
    #fn, closure_args = self.flatten_fn(expr.fn)
    #shape_elts = self.flatten_expr(expr.shape)
    # import pdb; pdb.set_trace()
    #return IndexMap(fn = self.closure(fn, closure_args), shape = self.tuple(shape_elts), type = None)
    assert False, "Unexpected IndexMap, should have been turned into ParFor before flattening"
    
  
  def flatten_OuterMap(self, expr):
    assert False, "Unexpected OuterMap, should have been turned into ParFor before flattening"
  
  def flatten_IndexReduce(self, expr):
    # assert isinstance(expr.type, ScalarT), "Non-scalar reductions not yet implemented"
    fn, fn_args = self.flatten_fn(expr.fn)
    fn = self.closure(fn, fn_args)
    combine, combine_args = self.flatten_fn(expr.combine)
    combine = self.closure(combine, combine_args)
    shape = self.tuple(self.flatten_expr(expr.shape))
    
    init = self.flatten_expr(expr.init)
    t = flatten_type(expr.type)
    if len(t) == 1:
      init = init[0]
      t = t[0]
    else:
      init = self.tuple(init)
      t = make_tuple_type(t)
    
    result =  IndexReduce(fn = fn, combine = combine, shape = shape, type = t, init = init)
    return [result]
     
      
  def flatten_IndexScan(self, expr):
    assert False, "IndexScan Not implemented" 
  
  def flatten_lhs_name(self, name, t):
    if isinstance(t, (ScalarT, PtrT)):
      return [Var(name = name, type = t)]
    elif isinstance(t, (NoneT, FnT, TypeValueT)):
      return []
    elif isinstance(t, SliceT):
      base = name.replace(".", "_")
      start = Var(name = "%s_start" % base, type = t.start_type)
      stop = Var(name = "%s_stop" % base, type = t.stop_type)
      step = Var(name = "%s_step" % base, type = t.step_type)
      field_vars = [start, stop, step]
    elif isinstance(t, ClosureT):
      base = name.replace(".", "_")
      field_vars = [Var(name = "%s_closure_elt%d" % (base,i) , type = t) 
                    for i,t in enumerate(t.arg_types)]
    elif isinstance(t, TupleT):
      base = name.replace(".", "_")
      field_vars = [Var(name = "%s_elt%d" % (base, i), type = t) 
                    for i,t in enumerate(t.elt_types)]
    elif isinstance(t, ArrayT):
      base = name.replace(".", "_")
      data = Var(name = "%s_data" % base, type = t.ptr_t)
      shape = Var(name = "%s_shape" % base, type = t.shape_t)
      strides = Var(name = "%s_strides" % base, type = t.strides_t)
      offset = Var(name = "%s_offset" % base, type = Int64)
      nelts = Var(name = "%s_nelts" % base, type = Int64)
      field_vars = [data, shape, strides, offset, nelts]
    else:
      assert False, "Unsupport type %s" % (t,)
    return self.flatten_lhs_vars(field_vars)
  
  def flatten_lhs_var(self, old_var):
    t = old_var.type 
    if isinstance(t, (PtrT, ScalarT)):
      return [old_var]
    name = old_var.name 
    return self.flatten_lhs_name(name, t)
  
  def flatten_lhs_vars(self, old_vars):
    return concat_map(self.flatten_lhs_var, old_vars)
  
  def flatten_scalar_lhs_var(self, old_var):
    lhs_vars = self.flatten_lhs_var(old_var)
    assert len(lhs_vars) == 1
    return lhs_vars[0]
  
def build_flat_fn(old_fn, _cache = {}):
  key = old_fn.cache_key
  if key in _cache:
    return _cache[key]
  flat_fn = BuildFlatFn(old_fn).run()
  _cache[key] = flat_fn
  _cache[flat_fn.cache_key] = flat_fn
  return flat_fn
  
class Flatten(Transform):
  
  def unbox_var(self, var):
    t = var.type 

    if isinstance(t, (FnT, NoneT, TypeValueT)):
      return []
    elif isinstance(t, (PtrT, ScalarT)):
      return [var]
    elif isinstance(t, ArrayT):
      # for structured arrays, should this be a tuple? 
      base = var.name.replace(".", "_")
      data = self.attr(var, 'data', name = base + "_data")
      shape = self.attr(var, 'shape', name = base + "_shape")
      strides = self.attr(var, 'strides', name = base + "_strides")
      offset = self.attr(var, 'offset', name = base + "_offset")
      size = self.attr(var, 'size', name = base + "_size")
      return self.unbox_vars([data, shape, strides, offset, size])
    elif isinstance(t, SliceT):
      start = self.attr(var, 'start')
      stop = self.attr(var, 'stop')
      step = self.attr(var, 'step')
      return self.unbox_vars([start, stop, step])
    elif isinstance(t, ClosureT):
      base = var.name.replace(".", "_")
      closure_elts = [self.closure_elt(var, i, name = base + "_closure_elt%d" % i)
                      for i in xrange(len(t.arg_types))]
      return self.unbox_vars(closure_elts)
    elif isinstance(t, TupleT):
      base = var.name.replace(".", "_")
      tuple_elts = [self.assign_name(self.tuple_proj(var, i), name = base + "_elt%d" % i)
                      for i in xrange(len(t.elt_types))]
      return self.unbox_vars(tuple_elts)
    else:
      assert False, "Unsupported type %s" % (t,)
  
  def unbox_vars(self, exprs):
    return concat_map(self.unbox_var, exprs)
  
  def to_seq(self, expr):
    if isinstance(expr.type, TupleT):
      return self.tuple_elts(expr)
    elif isinstance(expr.type, (FnT, NoneT)):
      return []
    else:
      return [expr]
  
  def box(self, t, elts):
    if isinstance(t, NoneT):
      assert len(elts) == 0, "Expected 0 values for None, got %s" % (elts,)
      return none 
    elif isinstance(t, ScalarT):
      assert len(elts) == 1
      return elts[0]
    elif isinstance(t, SliceT):
      assert len(elts) == 3
      start, stop, step = elts
      return self.slice_value(start, stop, step)
    elif isinstance(t, ArrayT):
      data = get_field_elts(t, elts, 'data')[0]
      shape = self.tuple(get_field_elts(t, elts, 'shape'))
      strides = self.tuple(get_field_elts(t, elts, 'strides'))
      offset = get_field_elts(t, elts, 'offset')[0]
      nelts = get_field_elts(t, elts, 'size')[0]
      return self.array_view(data, shape, strides, offset, nelts)
    elif isinstance(t, TupleT):
      boxed_elts = []
      for i, elt_t in enumerate(t.elt_types):
        elt = self.box(elt_t, get_field_elts(t, elts, i))
        boxed_elts.append(elt)
      return self.tuple(boxed_elts)
    elif isinstance(t, ClosureT):
      assert False, "Not implemented: ClosureT" 
    elif isinstance(t, FnT):
      assert False, "Not implemented: FnT" 
  
  def transform_block(self, stmts):
    return stmts
    
  def pre_apply(self, old_fn, _cache = {}):

    key = old_fn.cache_key
    if key  in _cache:
      return _cache[key]
    
    flat_fn = build_flat_fn(old_fn)
    
    flat_fn.created_by = old_fn.created_by
    flat_fn.transform_history = old_fn.transform_history.copy()
    
    input_vars = mk_vars(old_fn.arg_names, old_fn.input_types)
    self.blocks.push()
    unboxed_inputs = self.unbox_vars(input_vars)
    assert len(unboxed_inputs) == len(flat_fn.input_types)
    unboxed_result = self.call(flat_fn, unboxed_inputs, name = "unboxed_result")
    unboxed_elts = self.to_seq(unboxed_result)
    boxed_result = self.box(old_fn.return_type, unboxed_elts)
    self.return_(boxed_result)
    old_fn.body = self.blocks.pop()
    _cache[key] = old_fn
    _cache[flat_fn.cache_key] = flat_fn 
    return old_fn
  


import shape
from shape import * 
 
from shape_from_type import shapes_from_types
from shape_inference import (shape_env, call_shape_expr, bind, bind_pairs, 
                             subst, subst_list, symbolic_call)



from .. import names
from ..ndtypes import (StructT, Type, Unknown,
                       ArrayT, TypeValueT, TupleT,
                       ScalarT, IntT, Int64,  Bool, Float64,    
                       combine_type_list, increase_rank, lower_rank,  
                       make_array_type, make_tuple_type, make_slice_type, make_closure_type, 
                       repeat_tuple, 
                       type_conv, )

from ..syntax import (Array, AllocArray, Attribute, 
                      Cast, Closure, Const, ConstArray, ConstArrayLike, 
                      Expr, Index, 
                      Range, Ravel, Reshape, Shape,
                      Select, Slice, 
                      Transpose, Tuple, TupleProj, TypeValue,  Var,  
                      ForLoop, While, Assign, Return, If)

from ..syntax.helpers import get_types, is_true, is_false, const_int
from ..syntax.wrappers import build_untyped_prim_fn

from ..transforms import Transform


class LocalTypeInference(Transform):
  """
  Local type inference information which doesn't 
  require recursive calls back into the inference 
  algorithm. 
  """
  def __init__(self, tenv, var_map):
    Transform.__init__(self)
    self.type_env = tenv 
    self.var_map = var_map
    
  def transform_expr(self, expr):
    from ..frontend import ast_conversion
    if not isinstance(expr, Expr):
      expr = ast_conversion.value_to_syntax(expr)
    result = Transform.transform_expr(self, expr)  
    assert result.type is not None, \
       "Unsupported expression encountered during type inference: %s" % (expr,) 
    return result 


  def transform_Index(self, expr):
    value = self.transform_expr(expr.value)
    index = self.transform_expr(expr.index)
    if isinstance(value.type, TupleT):
      assert isinstance(index.type, IntT)
      assert index.__class__  is Const
      i = index.value
      assert isinstance(i, int)
      elt_types = value.type.elt_types
      assert i < len(elt_types), \
          "Can't get element %d of length %d tuple %s : %s" % \
          (i, len(elt_types), value, value.type)
      elt_t = value.type.elt_types[i]
      return TupleProj(value, i, type = elt_t)
    else:
      result_type = value.type.index_type(index.type)
      return Index(value, index, type = result_type)

  
  def transform_Array(self, expr):
    
    new_elts = self.transform_args(expr.elts)
    if expr.type is None:
      elt_types = get_types(new_elts)
      if len(elt_types) > 0:
        common_t = combine_type_list(elt_types)
        if common_t is Unknown:
          raise TypeError("Couldn't find commom type for elements of array %s" % expr)
      else:
        # numpy defaults to Float arrays 
        common_t = Float64 
    else:
      common_t = expr.type 
    array_t = increase_rank(common_t, 1)
    return Array(new_elts, type = array_t)

  def transform_AllocArray(self, expr):
    elt_type = expr.elt_type
    assert isinstance(elt_type, ScalarT), \
      "Invalid array element type  %s" % (elt_type)
      
    shape = self.transform_expr(expr.shape)
    if isinstance(shape, ScalarT):
      shape = self.cast(shape, Int64)
      shape = self.tuple((shape,), "array_shape")
    assert isinstance(shape, TupleT), \
      "Invalid shape %s" % (shape,)
    rank = len(shape.elt_types)
    t = make_array_type(elt_type, rank)
    return AllocArray(shape, elt_type, type = t)

  def transform_ConstArray(self, expr):
    shape = self.transform_expr(expr.shape)
    value = self.transform_expr(expr.value)
    assert isinstance(value.type, ScalarT), \
      "ConstArray expects scalar value, got %s: %s" % (value, value.type)
    if isinstance(shape.type, ScalarT):
      shape = self.tuple([shape])
    assert isinstance(shape.type, TupleT), \
      "Expected shape of ConstArray to be tuple, got %s : %s" % (shape, shape.type)
    ndims = len(shape.type.elt_types)
    if ndims == 0:
      return value 
    array_t = make_array_type(value.type, ndims)
    return ConstArray(shape, value, type = array_t) 
  
  def transform_ConstArrayLike(self, expr):
    array = self.transform_expr(expr.array)
    value = self.transform_expr(expr.value)
    assert isinstance(value.type, ScalarT), \
      "ConstArray expects scalar value, got %s: %s" % (value, value.type)
    if isinstance(array.type, ScalarT):
      return value 
    elif isinstance(array.type, TupleT):
      array = self.transform_expr(Array(elts = self.tuple_elts(array)))
      
    assert isinstance(array.type, ArrayT), \
      "ConstArrayLike expected array, got %s : %s" % (array, array.type)
    shape = self.shape(array)
    ndims = len(shape.type.elt_types)
    if ndims == 0:
      return value 
    array_t = make_array_type(value.type, ndims)
    return ConstArrayLike(array = array, value = value, type = array_t)
    

  
  def transform_Range(self, expr):
    start = self.transform_expr(expr.start) if expr.start else None
    stop = self.transform_expr(expr.stop) if expr.stop else None
    step = self.transform_expr(expr.step) if expr.step else None
    
    # by default we're generating ranges of Int64
    # but also allow for floating values 
    elt_t = Int64 
    if not self.is_none(start): elt_t = elt_t.combine(start.type)
    if not self.is_none(stop): elt_t = elt_t.combine(stop.type)
    if not self.is_none(step): elt_t = elt_t.combine(step.type)
    array_t = make_array_type(elt_t, 1)
    return Range(start, stop, step, type = array_t)

  def transform_Slice(self, expr):
    start = self.transform_expr(expr.start)
    stop = self.transform_expr(expr.stop)
    step = self.transform_expr(expr.step)
    slice_t = make_slice_type(start.type, stop.type, step.type)
    return Slice(start, stop, step, type = slice_t)


  def transform_Var(self, expr):
    old_name = expr.name
    if old_name not in self.var_map._vars:
      raise names.NameNotFound(old_name)
    new_name = self.var_map.lookup(old_name)
    
    assert new_name in self.type_env, \
        "Unknown var %s (previously %s)" % (new_name, old_name)
    t = self.type_env[new_name]
    return Var(new_name, type = t)

  def transform_Tuple(self, expr):
    elts = self.transform_expr_list(expr.elts)
    return self.tuple(elts)

  def transform_Const(self, expr):
    if expr.type is None:
      t = type_conv.typeof(expr.value)
    else:
      t = expr.type 
    return Const(expr.value, type = t)
  

  def transform_Shape(self, expr):
    array = self.transform_expr(expr.array)
    if isinstance(array.type, ArrayT):
      t = repeat_tuple(Int64, array.type.rank)
      return Shape(array = array, type = t)
    elif isinstance(array.type, TupleT):
      elt = const_int(len(array.type.elt_types))
      return self.tuple( [elt]  )
    else:
      return self.tuple( [] )
      

  def transform_Reshape(self, expr):
    array = self.transform_expr(expr.array)
    shape = self.transform_expr(expr.shape)
    rank = len(shape.type.elt_types)
    assert array.type.__class__ is ArrayT
    t = make_array_type(array.elt_type, rank)
    return Reshape(array, shape, type = t)
  
  def transform_Ravel(self, expr):
    array = self.transform_expr(expr.array)
    if isinstance(array.type, ScalarT):
      return array 
    assert array.type.__class__ is ArrayT, \
      "Can't ravel/flatten %s of type %s" % (array, array.type) 
    t = make_array_type(array.type.elt_type, 1)
    return Ravel(array, type = t)
  
  def transform_Transpose(self, expr):
    array = self.transform_expr(expr.array)
    if isinstance(array.type, ScalarT):
      return array
    assert array.type.__class__ is ArrayT, \
      "Can't transpose %s of type %s" % (array, array.type) 
    return Transpose(array = array, type = array.type)
  
  def transform_Cast(self, expr):
    v = self.transform_expr(expr.value)
    return Cast(v, type = expr.type)
  
  def transform_Len(self, expr):
    v = self.transform_expr(expr.value)
    t = v.type
    if t.__class__ is ArrayT:
      shape_t = make_tuple_type([Int64] * t.rank)
      shape = Attribute(v, 'shape', type = shape_t)
      return TupleProj(shape, 0, type = Int64)
    else:
      assert t.__class__ is TupleT, "Unexpected argument for 'len' - %s : %s" % (expr.value, t)
      return Const(len(t.elt_types), type = Int64)

  def transform_DelayUntilTyped(self, expr):
    new_values = self.transform_expr_tuple(expr.values)
    if expr.keywords:
      typed_keywords = {}
      for k, v in expr.keywords.iteritems():
        typed_keywords[k] = self.transform_expr(v)
      new_syntax = expr.fn(*new_values, **typed_keywords)
    else:
      new_syntax = expr.fn(*new_values)
    assert new_syntax.type is not None, \
      "Error in %s, new expression %s lacks type" % (expr, new_syntax)
    return new_syntax
  
  def transform_TypeValue(self, expr):
    t = expr.type_value 
    assert isinstance(t, Type), "Invalid type value %s" % (t,)
    return TypeValue(t, type=TypeValueT(t))
    
  def transform_Closure(self, expr):
    new_args = self.transform_expr_list(expr.args)
    t = make_closure_type(expr.fn, get_types(new_args))
    return Closure(expr.fn, new_args, type = t)

  def transform_Arith(self, expr):
    return build_untyped_prim_fn(expr)
    #t = make_closure_type(untyped_fn, ())
    #return Closure(untyped_fn, (), type = t)

  def transform_UntypedFn(self, expr):
    return expr 

  def transform_Attribute(self, expr):
    value = self.transform_expr(expr.value)
    assert isinstance(value.type, StructT)
    result_type = value.type.field_type(expr.name)
    return Attribute(value, expr.name, type = result_type)
  
  def transform_Select(self, expr):
    trueval = self.transform_expr(expr.true_value)
    falseval = self.transform_expr(expr.false_value)
    t = trueval.type.combine(falseval.type)
    trueval = self.cast(trueval, t)
    falseval = self.cast(falseval, t) 
    cond = self.transform_expr(expr.cond)
    cond = self.cast(cond, Bool)
    return Select(cond = cond, true_value = trueval, false_value = falseval, type = t)
  
  def infer_phi(self, result_var, val):
    """
    Don't actually rewrite the phi node, just add any necessary types to the
    type environment
    """

    new_val = self.transform_expr(val)
    new_type = new_val.type
    old_type = self.type_env.get(result_var, Unknown)
    new_result_var = self.var_map.lookup(result_var)
    self.type_env[new_result_var]  = old_type.combine(new_type)

  def infer_phi_nodes(self, nodes, direction):
    for (var, values) in nodes.iteritems():
      self.infer_phi(var, direction(values))

  def infer_left_flow(self, nodes):
    return self.infer_phi_nodes(nodes, lambda (x, _): x)

  def infer_right_flow(self, nodes):
    return self.infer_phi_nodes(nodes, lambda (_, x): x)

  
  def transform_phi_node(self, result_var, (left_val, right_val)):
    """
    Rewrite the phi node by rewriting the values from either branch, renaming
    the result variable, recording its new type, and returning the new name
    paired with the annotated branch values
    """

    new_left = self.transform_expr(left_val)
    new_right = self.transform_expr(right_val)
    old_type = self.type_env.get(result_var, Unknown)
    new_type = old_type.combine(new_left.type).combine(new_right.type)
    new_var = self.var_map.lookup(result_var)
    self.type_env[new_var] = new_type
    return (new_var, (new_left, new_right))

  def transform_phi_nodes(self, nodes):
    new_nodes = {}
    for old_k, (old_left, old_right) in nodes.iteritems():
      new_name, (left, right) = self.transform_phi_node(old_k, (old_left, old_right))
      new_nodes[new_name] = (left, right)
    return new_nodes

  def annotate_lhs(self, lhs, rhs_type):

    lhs_class = lhs.__class__
    #
    # Am I a tuple of Assignments? 
    #
    if lhs_class is Tuple:
      if rhs_type.__class__ is TupleT:
        assert len(lhs.elts) == len(rhs_type.elt_types)
        new_elts = [self.annotate_lhs(elt, elt_type) 
                    for (elt, elt_type) 
                    in zip(lhs.elts, rhs_type.elt_types)]
      else:
        assert rhs_type.__class__ is ArrayT, \
            "Unexpected right hand side type %s for %s" % (rhs_type, lhs)
        elt_type = lower_rank(rhs_type, 1)
        new_elts = [self.annotate_lhs(elt, elt_type) for elt in lhs.elts]
      tuple_t = make_tuple_type(get_types(new_elts))
      return Tuple(new_elts, type = tuple_t)
    
    #
    # Index Assignment
    #
    elif lhs_class is Index:
      new_arr = self.transform_expr(lhs.value)
      new_idx = self.transform_expr(lhs.index)
      assert new_arr.type.__class__ is ArrayT, \
          "Expected array, got %s" % new_arr.type
      elt_t = new_arr.type.index_type(new_idx.type)
      return Index(new_arr, new_idx, type = elt_t)
    
    # 
    # Attribute Assignment for Mutable Objects
    # 
    elif lhs_class is Attribute:
      name = lhs.name
      struct = self.transform_expr(lhs.value)
      struct_t = struct.type
      assert isinstance(struct_t, StructT), \
          "Can't access fields on value %s of type %s" % \
          (struct, struct_t)
      field_t = struct_t.field_type(name)
      return Attribute(struct, name, field_t)
    
    # 
    # Regular Binding of Names to Values
    # 
    else:
      assert lhs_class is Var, "Unexpected LHS: %s" % (lhs,)
      new_name = self.var_map.lookup(lhs.name)
      old_type = self.type_env.get(new_name, Unknown)
      new_type = old_type.combine(rhs_type)
      self.type_env[new_name] = new_type
      return Var(new_name, type = new_type)

  def transform_Assign(self, stmt):
    rhs = self.transform_expr(stmt.rhs)
    lhs = self.annotate_lhs(stmt.lhs, rhs.type)
    return Assign(lhs, rhs)
  
  def transform_If(self, stmt):

    cond = self.transform_expr(stmt.cond) 

    assert isinstance(cond.type, ScalarT), \
        "Condition %s has type %s but must be convertible to bool" % (cond, cond.type)
    # it would be cleaner to not have anything resembling an optimization 
    # inter-mixed with the type inference, but I'm not sure how else to 
    # support 'if x is None:...'
    if is_true(cond):
      self.blocks.top().extend(self.transform_block(stmt.true))
      for (name, (left,_)) in stmt.merge.iteritems():
        typed_left = self.transform_expr(left)
        typed_var = self.annotate_lhs(Var(name), typed_left.type) 
        self.assign(typed_var, typed_left)
      return
    
    if is_false(cond):
      self.blocks.top().extend(self.transform_block(stmt.false))
      for (name, (_,right)) in stmt.merge.iteritems():
        typed_right = self.transform_expr(right)
        typed_var = self.annotate_lhs(Var(name), typed_right.type)
        self.assign(typed_var, typed_right)
      return
    true = self.transform_block(stmt.true)
    false = self.transform_block(stmt.false) 
    merge = self.transform_phi_nodes(stmt.merge)
    return If(cond, true, false, merge)

  def transform_Return(self, stmt):
    ret_val = self.transform_expr(stmt.value)
    curr_return_type = self.type_env["$return"]
    self.type_env["$return"] = curr_return_type.combine(ret_val.type)
    return Return(ret_val)

  def transform_While(self, stmt):
    self.infer_left_flow(stmt.merge)
    cond = self.transform_expr(stmt.cond)
    body = self.transform_block(stmt.body)
    merge = self.transform_phi_nodes(stmt.merge)
    return While(cond, body, merge)

  def transform_ForLoop(self, stmt):
    start = self.transform_expr(stmt.start)
    stop = self.transform_expr(stmt.stop)
    step = self.transform_expr(stmt.step)
    lhs_t = start.type.combine(stop.type).combine(step.type)
    var = self.annotate_lhs(stmt.var, lhs_t)
    self.infer_left_flow(stmt.merge)
    body = self.transform_block(stmt.body)
    merge = self.transform_phi_nodes(stmt.merge)
    return ForLoop(var, start, stop, step, body, merge)




import numpy as np 


def smooth(x, alpha):
  s = x.copy()
  for i in xrange(1, len(x)):
      s[i] = alpha * x[i] + (1-alpha)*s[i-1]
  return s 

n = 10**6
alpha = 0.01
X = np.random.randn(n).astype('float32')

from compare_perf import compare_perf
compare_perf(smooth, [X, alpha])




from core_types import  IncompatibleTypes, ImmutableT

class FnT(ImmutableT):
  """Type of a typed function"""
  def __init__(self, input_types, return_type):
    self.input_types = tuple(input_types)
    self.return_type = return_type 
    self._hash = hash(self.input_types + (return_type,))

  def __str__(self):
    input_str = ", ".join(str(t) for t in self.input_types)
    return "(%s)->%s" % (input_str, self.return_type)

  def __repr__(self):
    return str(self)

  def __eq__(self, other):
    return other.__class__ is FnT and \
           self.return_type == other.return_type and \
           len(self.input_types) == len(other.input_types) and \
           all(t1 == t2 for (t1, t2) in
               zip(self.input_types, other.input_types))

  def combine(self, other):
    if self == other:
      return self
    else:
      raise IncompatibleTypes(self, other)

  def __hash__(self):
    return self._hash

_fn_type_cache = {}
def make_fn_type(input_types, return_type):
  input_types = tuple(input_types)
  key = input_types, return_type
  if key in _fn_type_cache:
    return _fn_type_cache[key]
  else:
    t = FnT(input_types, return_type)
    _fn_type_cache[key] = t
    return t


from .. import syntax
from .. syntax import Const, ForLoop, Var
from .. syntax.helpers import const_int
from ..transforms import CloneStmt 

from loop_transform import LoopTransform


def safediv(m,n):
  return (m+n-1)/n

class LoopUnrolling(LoopTransform):
  def __init__(self, unroll_factor = 4,
                      max_static_unrolling = 8,
                      max_block_size = 50):
    LoopTransform.__init__(self)
    self.unroll_factor = unroll_factor
    if max_static_unrolling is not None:
    # should we unroll static loops more than ones with unknown iters?
      self.max_static_unrolling = max_static_unrolling
    else:
      self.max_static_unrolling = unroll_factor

    self.max_block_size = max_block_size

  def pre_apply(self, fn):
    # skip the alias analysis that's default for LoopTransform
    return fn

  def copy_loop_body(self, stmt, outer_loop_var, iter_num, phi_values = None):
    """Assume the current codegen block is the unrolled loop"""

    cloner = CloneStmt(self.type_env)
    # make a fresh copy of the loop
    loop = cloner.transform_ForLoop(stmt)
    i = const_int(iter_num, loop.var.type)
    loop_var_value = self.add(outer_loop_var, self.mul(stmt.step, i))
    self.assign(loop.var, loop_var_value)

    # if this isn't the first iteration of unrolling
    # then propagate old versions of phi-bound values
    # into this block
    if phi_values is not None:
      for (old_name, input_value) in phi_values.iteritems():
        new_var = cloner.rename_dict[old_name]
        self.assign(new_var, input_value)

    self.blocks.top().extend(loop.body)

    output_values = {}
    for old_name in stmt.merge.iterkeys():
        new_var = cloner.rename_dict[old_name]
        output_values[old_name] = loop.merge[new_var.name][1]

    return output_values, cloner.rename_dict

  def transform_ForLoop(self, stmt):
    assert self.unroll_factor > 0
    if self.unroll_factor == 1:
      return stmt

    if stmt.step.__class__ is Const:
      assert stmt.step.value > 0, "Downward loops not yet supported"

    stmt = LoopTransform.transform_ForLoop(self, stmt)

    if not self.is_simple_block(stmt.body) or \
       len(stmt.body) > self.max_block_size:
      return stmt

    start, stop, step = stmt.start, stmt.stop, stmt.step

    # if loop has static bounds, fully unroll unless it's too big
    unroll_factor = self.unroll_factor

    # number of iterations of loop iterations is not generally known
    if start.__class__ is Const and \
       stop.__class__ is Const and \
       step.__class__ is Const:
      niters = safediv(stop.value - start.value, step.value)
      if niters <= self.max_static_unrolling:
        unroll_factor = niters

    # push the unrolled body onto the stack
    self.blocks.push()

    phi_values = None
    loop_var = self.fresh_var(stmt.var.type,  "loop_counter")
    name_mappings = None
    for iter_num in xrange(unroll_factor):
      #self.comment("Unrolling iteration %d" % iter_num)
      phi_values, curr_names = \
          self.copy_loop_body(stmt, loop_var, iter_num, phi_values)
      if name_mappings is None:
        name_mappings = curr_names

    unrolled_body = self.blocks.pop()
    unroll_value = syntax.helpers.const_int(unroll_factor, stmt.var.type)
    unrolled_step = self.mul(unroll_value, stmt.step, "unrolled_step")
    n_total_iters = self.sub(stop, start, name = "niters")
    n_unrolled_iters = self.div(n_total_iters, unrolled_step, name = "unrolled_iters")
    # Python's division doesn't behave like C, so that small_negative/big_positive = -1
    # ..which is crappy when expecting truncation! 
    n_unrolled_iters = self.max(n_unrolled_iters, self.int(0), name = "unrolled_iters")
    trunc = self.mul(n_unrolled_iters, unrolled_step, "trunc")
    unrolled_stop = self.add(stmt.start, trunc, "unrolled_stop")
    final_merge = {}
    for (old_name, (input_value, _)) in stmt.merge.iteritems():
      first_name_in_loop = name_mappings[old_name].name
      output_value = phi_values[old_name]
      final_merge[first_name_in_loop] = (input_value, output_value)

    unrolled_loop = ForLoop(var = loop_var,
                            start = stmt.start,
                            stop = unrolled_stop,
                            step = unrolled_step,
                            body = unrolled_body,
                            merge = final_merge)

    if unrolled_loop.start.__class__ is Const and \
       unrolled_loop.stop.__class__ is Const and \
       unrolled_loop.step.__class__ is Const:
      start_value = unrolled_loop.start.value
      stop_value = unrolled_stop.value
      step_value = unrolled_loop.step.value
      if start_value + step_value == stop_value:
        self.assign(unrolled_loop.var, unrolled_loop.start)
        # assign all loop-carried variables to their initial values
        if len(final_merge) > 0:
          self.comment("Initialize loop-carried values")
        for (acc_name, (input_value, _)) in final_merge.iteritems():
          var = Var(acc_name, type = input_value.type)
          self.assign(var, input_value)
        # inline loop body
        self.blocks.top().extend(unrolled_body)
        # since we're not going to have a cleanup loop,
        # need to assign all the original phi-carried variables
        if len(stmt.merge) > 0:
          self.comment("Finalize loop-carried values")
        for old_acc_name in stmt.merge.iterkeys():
          last_value = phi_values[old_acc_name]
          var = Var(old_acc_name, last_value.type)
          self.assign(var, last_value)
        return None

    self.blocks.append(unrolled_loop)

    if unrolled_loop.stop.__class__ is not Const or \
       stop.__class__ is not Const or \
       unrolled_loop.stop.value != stop.value:
      cleanup_merge = {}
      for (old_name, (_, output_value)) in stmt.merge.iteritems():
        input_var = name_mappings[old_name]
        cleanup_merge[old_name] = (input_var, output_value)
      stmt.merge = cleanup_merge
      stmt.start = unrolled_loop.stop
    return stmt





from .. analysis.index_elim_analysis import IndexElimAnalysis
from .. analysis.index_elim_analysis import \
  (unknown, ConstValue, ConstElts, RangeArray, IndexMapResult)
from .. ndtypes import ScalarT, TupleT 
from .. syntax import Const, Var 
from transform import Transform   
 

        
class IndexElim(Transform):
  def pre_apply(self, fn):
    analysis = IndexElimAnalysis() 
    analysis.visit_fn(fn)
    self.array_values = analysis.array_values
     
  def transform_Index(self, expr):
    if expr.value.__class__ is not Var: return expr 
    x = expr.value.name 
    if x not in self.array_values: return expr
    v = self.array_values[x]
    if v is unknown: return expr 
    if self.is_tuple(expr.index):
      indices = self.tuple_elts(expr.index)
    else:
      indices = [expr.index]
    if not all(isinstance(idx.type, ScalarT) for idx in indices): return expr
     
    n_indices = len(indices)
    if v.__class__ is ConstValue:
      if n_indices != v.type.rank: return expr
      return Const(v.value, v.type)
    elif v.__class__ is ConstElts  and all(idx.__class__ is Const for idx in indices):
      if n_indices != v.type.rank: return expr
      idx_values = [idx.value for idx in indices]
      return Const(v.array[tuple(idx_values)], type = v.type.elt_type)
    elif v.__class__ is IndexMapResult: 
      fn = self.get_fn(v.fn)
      closure_args = self.get_closure_args(v.fn)
      input_types = fn.input_types
      last_input_type =  input_types[-1]
      if last_input_type.__class__ is TupleT:
        n_expected_indices = len(last_input_type.elt_types)
        indices_are_tuple = True 
      else:
        indices_are_tuple = False
        n_expected_indices = len(input_types) - len(closure_args)
      if n_indices != n_expected_indices:
        return expr 
      if indices_are_tuple:
        index_exprs = [self.tuple(indices)]
      else:
        index_exprs = list(indices)
      return self.call(v.fn, index_exprs)
        
    assert v.__class__ is RangeArray
    assert len(indices) == 1
    idx = indices[0]
    if idx.__class__ is Const and v.step.__class__ is Const and v.start.__class__ is Const:
      return Const(idx.value * v.step.value + v.start.value, type = v.type) 

    return self.add(v.start, self.mul(idx, v.step))
      


import numpy as np 

from parakeet.testing_helpers import run_local_tests, expect 


int_vec = np.array([2,3])
int_mat = np.array([int_vec, int_vec])

bool_vec = np.array([True, True])
bool_mat = np.array([bool_vec, bool_vec])

float32_vec = np.array([1.0, 2.0], dtype='float32')
float32_mat = np.array([float32_vec, float32_vec])

float64_vec = np.array([1.0, 10.0])
float64_mat = np.array([float64_vec, float64_vec])

vecs = [int_vec, bool_vec, float32_vec, float64_vec]
mats = [int_mat, bool_mat, float32_mat, float64_mat]

def binary(fn):
  for x in mats:
    for y in mats:
      expect(fn, [x,y], fn(x,y), x.dtype)
      
def test_add():
  binary(np.add)

def test_subtract():
  binary(np.subtract)

def test_multiply():
  binary(np.multiply)
  
def test_divide():
  binary(np.divide)
  
def test_logaddexp():
  binary(np.logaddexp)
  
def test_logaddexp2():
  binary(np.logaddexp2)

def test_true_divide():
  binary(np.true_divide)
  
def test_floor_divide():
  binary(np.floor_divide)
def test_power():
  binary(np.power)

def test_remainder():
  binary(np.remainder)
  
def test_mod():
  binary(np.mod)
  
def test_fmod():
  binary(np.fmod)
  
      
if __name__ == '__main__':
  run_local_tests() 
  




from parakeet.testing_helpers import expect, run_local_tests

def volatile_licm_mistake():
  i = 0
  x = [0]
  while i < 10:
    alloc = [1]
    if i == 5:
      x = alloc
    else:
      alloc[0] = 100
    i = i + 1
  return x[0]

def test_volatile_licm_mistake():
  expect(volatile_licm_mistake, [], volatile_licm_mistake()) #np.array([1]))

def licm_nested_loops():
  
  total = 0
  a = [1,2,3,4,5]
  for _ in range(3):
    for _ in range(2):
      for j in range(len(a)):
        a[j] *= 10
    total += a[1]
  return total   

def test_licm_nested_loops():
  expect(licm_nested_loops, [], licm_nested_loops())

def loop_invariant_alloc():
  b = [1,2,3]
  total = 0
  for i in xrange(3):
    a = [0,0,0]
    for j in xrange(3):
      a[j] = b[j]
    total += a[i]
  return total 
  
def test_loop_invariant_alloc():
  expect(loop_invariant_alloc, [], loop_invariant_alloc())
  

if __name__ == "__main__":
  run_local_tests()


from .. import config 
from ..analysis import (contains_adverbs, contains_calls, contains_loops, 
                        contains_structs, contains_array_operators)

from combine_nested_maps import CombineNestedMaps 
from copy_elimination import CopyElimination
from dead_code_elim import DCE

from flattening import Flatten
from fusion import Fusion
from imap_elim import IndexMapElimination
from index_elimination import IndexElim
from indexify_adverbs import IndexifyAdverbs

from inline import Inliner
from licm import LoopInvariantCodeMotion
from loop_unrolling import LoopUnrolling
from lower_adverbs import LowerAdverbs
from lower_array_operators import LowerArrayOperators
from lower_indexing import LowerIndexing
from lower_slices import LowerSlices
from lower_structs import LowerStructs
from negative_index_elim import NegativeIndexElim
from offset_propagation import OffsetPropagation
from parfor_to_nested_loops import ParForToNestedLoops
from phase import Phase
from range_propagation import RangePropagation
from redundant_load_elim import RedundantLoadElimination
from scalar_replacement import ScalarReplacement
from shape_elim import ShapeElimination
from simplify import Simplify
from simplify_array_operators import SimplifyArrayOperators
from specialize_fn_args import SpecializeFnArgs

####################################
#                                  #
#    HIGH LEVEL OPTIMIZATIONS      # 
#                                  #
####################################

normalize = Phase([Simplify], 
                  memoize = True, 
                  copy = False, 
                  cleanup = [], 
                  name = "Normalize")

fusion_opt = Phase(Fusion, 
                   config_param = 'opt_fusion',
                   memoize = False,
                   copy = False, 
                   run_if = contains_adverbs)

inline_opt = Phase(Inliner, 
                   config_param = 'opt_inline', 
                   cleanup = [], 
                   run_if = contains_calls, 
                   memoize = False, 
                   copy = False)



licm = Phase(LoopInvariantCodeMotion, 
             config_param = 'opt_licm',
             run_if = contains_loops, 
             memoize = False, 
             name = "LICM")


symbolic_range_propagation = Phase([RangePropagation, OffsetPropagation], 
                                    config_param = 'opt_range_propagation',
                                    name = "SymRangeProp",
                                    copy = False,  
                                    memoize = False, 
                                    cleanup = [])
 
shape_elim = Phase(ShapeElimination,
                   config_param = 'opt_shape_elim')





early_optimizations = Phase([
                               inline_opt,
                               symbolic_range_propagation,   
                               licm,
                             ],
                            name = "EarlyOpt",
                            copy = True, 
                            memoize = True, 
                            cleanup = [Simplify, DCE], 
                            depends_on = normalize, 
 
                            )

simplify_array_operators = Phase([SimplifyArrayOperators], 
                                 copy = False, 
                                 memoize = False, 
                                 run_if = contains_array_operators, 
                                 config_param = "opt_simplify_array_operators")

combine_nested_maps = Phase([CombineNestedMaps],
                            copy = False, 
                            memoize = False, 
                            run_if = contains_adverbs, 
                            config_param = "opt_combine_nested_maps")

arg_specialization = Phase([SpecializeFnArgs],
                           copy = False, 
                           memoize = False, 
                           config_param = "opt_specialize_fn_args")

adverb_optimizations = Phase([
                                fusion_opt,
                                simplify_array_operators, 
                                combine_nested_maps,
                                arg_specialization,
                                fusion_opt, 
                                arg_specialization,
                              ], 
                             run_if = contains_adverbs, 
                             depends_on = early_optimizations, 
                             copy = True, 
                             memoize = True, 
                             cleanup = [Simplify, DCE])

high_level_optimizations = Phase([ 
                                    LowerArrayOperators,
                                     
                                    NegativeIndexElim,
                                    shape_elim, 
                                    symbolic_range_propagation, 
                                    
                                 ], 
                                 depends_on = adverb_optimizations,
                                 name = "HighLevelOpts", 
                                 copy = True, 
                                 memoize = True, 
                                 cleanup = [Simplify, DCE])



copy_elim = Phase(CopyElimination, 
                  config_param = 'opt_copy_elimination', 
                  memoize = False)


def print_indexified(fn):
  if config.print_indexified_function:
    print
    print "=== Indexified function ==="
    print
    print repr(fn)
    print


indexify = Phase([
                    IndexifyAdverbs, Simplify, DCE, 
                 ],
                 name = "Indexify",  
                 run_if = contains_adverbs, 
                 depends_on= high_level_optimizations, 
                 copy = True,
                 memoize = True, 
                 post_apply = print_indexified) 

after_indexify = Phase([copy_elim, Simplify, DCE, 
                        LowerSlices, 
                        inline_opt, Simplify, DCE, 
                        IndexMapElimination], 
                       name = "AfterIndexify", 
                       depends_on = indexify, 
                       copy = True, 
                       memoize = True)


flatten = Phase([Flatten, inline_opt, Simplify, DCE ], name="Flatten", 
                depends_on=after_indexify,
                run_if = contains_structs,  
                copy=True, 
                memoize = True)

####################
#                  #
#     LOOPIFY      # 
#                  #
####################

def print_loopy(fn):
  if config.print_loopy_function:
    print
    print "=== Loopy function ==="
    print
    print repr(fn)
    print




# loop_fusion = Phase(LoopFusion, config_param = 'opt_loop_fusion')


index_elim = Phase([NegativeIndexElim, IndexElim], config_param = 'opt_index_elimination')



lower_adverbs = Phase([LowerAdverbs, LowerSlices], run_if = contains_adverbs)


# Currently incorrect in the presence of function calls
# TODO: Make this mark an slices that are call arguments as read & written to
load_elim = Phase(RedundantLoadElimination,
                  config_param = 'opt_redundant_load_elimination', 
                  run_if = lambda fn: not contains_calls(fn))

loopify = Phase([lower_adverbs,
                 ParForToNestedLoops,
                 inline_opt,
                 LowerSlices, 
                 licm, 
                 shape_elim, 
                 symbolic_range_propagation,
                 load_elim,  
                 index_elim, 
                ],
                depends_on = after_indexify,
                cleanup = [Simplify, DCE],
                copy = True,
                memoize = True, 
                name = "Loopify",
                post_apply = print_loopy)


####################
#                  #
#     LOWERING     # 
#                  #
####################



lowering = Phase([
                    LowerIndexing,
                    licm,
                    LowerStructs,
                 ],
                 depends_on = loopify,
                 memoize = True, 
                 copy = True,
                 name = "Lowering", 
                 cleanup = [Simplify, DCE])

############################
#                          #
#  FINAL LOOP OPTIMIZATINS #
#                          #
############################

scalar_repl = Phase(ScalarReplacement, 
                    config_param = 'opt_scalar_replacement', 
                    run_if = contains_loops)



unroll = Phase([LoopUnrolling, licm], 
               config_param = 'opt_loop_unrolling', 
               run_if = contains_loops)



final_loop_optimizations = Phase([
                              licm, 
                              unroll, 
                              load_elim, 
                              scalar_repl, 
                              Simplify
                            ],
                           cleanup = [Simplify, DCE], 
                           copy = False, 
                           memoize = True,\
                           name = "FinalLoopOptimizations"
                           )


from syntax_visitor import SyntaxVisitor

from .. ndtypes  import TupleT, StructT, Type, PtrT, ClosureT, FnT

class TypeBasedMutabilityAnalysis(SyntaxVisitor):
  """
  The cheapest approximation to discovering which 
  fields are mutable is to do an insensitive analysis that
  marks all types which are on the RHS of a field assignment
  or are passed to a function (since other functions might 
  perform arbitrary mutable actions on fields). 
  
  Enter into the analysis via visit_fn, which returns 
  a set of struct types which may be mutable. 
  """
  
  def __init__(self):
    SyntaxVisitor.__init__(self)
    self.mutable_types = set([])
  
  def _mark_type(self, t):
    if self._has_mutable_fields(t):
      self.mutable_types.add(t)
      self._mark_children(t)
    elif isinstance(t, PtrT):
      self.mutable_types.add(t)
      

  def _has_mutable_fields(self, t):
    if isinstance(t, StructT):
      return not isinstance(t, (TupleT, ClosureT, FnT))
    
  def _mark_children(self, t):
    """
    For any type other than a Tuple or Closure, try 
    marking all its children 
    """
    for elt_t in t.children():
      self._mark_type(elt_t)
 
  
  """    
  def _mark(self, obj):
    if isinstance(obj, Type):
      self._mark_type(obj)
    elif isinstance(obj, (tuple, list)):
      for child in obj:
        self._mark(child)
  """   
  def visit_merge(self, phi_nodes):
    pass 
    
  def visit_generic_expr(self, expr):
    pass 
  
  def visit_lhs_Tuple(self, expr):
    for e in expr.elts:
      self.visit_lhs(e)
  
  def visit_lhs_Attribute(self, expr):
    assert False, "Considering making attributes immutable"
    self._mark_type(expr.value.type)
  
  def visit_lhs_Index(self, expr):
 
    self._mark_type(expr.value.type)
  
  def visit_Call(self, expr):
    for arg in expr.args:
      """
      The fields of an argument type might change, 
      but since we pass by value the argument itself
      isn't made mutable
      """
      self._mark_type(arg.type)
    
  def visit_fn(self, fn):
    self.mutable_types.clear()
    self.visit_block(fn.body)
    return self.mutable_types

def find_mutable_types(fn):
  return TypeBasedMutabilityAnalysis().visit_fn(fn) 



from adverbs import * 

from adverb_helpers import * 

from actual_args import ActualArgs 

from array_expr import (AllocArray, Array, ArrayExpr,  ArrayView,
                        ConstArray, ConstArrayLike,
                        DiagonalArray, 
                        Range, Ravel, Reshape, 
                        Shape, Slice, Strides, 
                        Transpose)
  

from delay_until_typed import DelayUntilTyped

from expr import Attribute, Call, Cast, Const, Closure, ClosureElt, Expr 
from expr import PrimCall, Select, Var 

from formal_args import FormalArgs, MissingArgsError, TooManyArgsError

import helpers 
from helpers import * 

from low_level import Alloc, Struct, Free, SourceExpr, SourceStmt

from seq_expr import Index, Enumerate, Len, Zip 

from stmt import (Stmt, Assign, Comment, ExprStmt, ForLoop, If, Return, While, ParFor, 
                  PrintString, block_to_str) 

from source_info import SourceInfo
from tuple_expr import Tuple, TupleProj

from typed_fn import TypedFn 
from type_value import TypeValue

from untyped_fn import UntypedFn 

from wrappers import build_untyped_prim_fn, build_untyped_expr_fn, build_untyped_cast_fn




class CombinedIters:
  def __init__(self, i1, i2):
    self.i1 = i1
    self.i2 = i2
    self.first_iter = True

  def next(self):
    if self.first_iter:
      try:
        return self.i1.next()
      except:
        self.first_iter = False
        return self.i2.next()
    else:
      return self.i2.next()

def combine_iters(*iters):
  assert len(iters) > 0
  curr_iter = iters[0]
  for i in iters[1:]:
    curr_iter = CombinedIters(curr_iter, i)
  return curr_iter

class NullIter(object):
  def next(self):
    raise StopIteration

def maybe_iter(obj):
  if obj is None:
    return NullIter()
  else:
    return iter(obj)

class ActualArgs(object):
  __slots__ = ['positional', 'keywords', 'starargs', "_hash"]
  
  def __init__(self, positional, keywords = {}, starargs = None):
    positional = tuple(positional)
    self.positional = positional 
    self.keywords = keywords
    self.starargs = starargs
    self._hash = None 
    
  def transform(self, fn, keyword_name_fn = None, keyword_value_fn = None):
    new_pos = [fn(pos_arg) for pos_arg in self.positional]
    new_keywords = {}
    for (k,v) in self.keywords.iteritems():
      new_name = keyword_name_fn(k) if keyword_name_fn else k
      new_value = keyword_value_fn(v) if keyword_value_fn else fn(v)
      new_keywords[new_name] = new_value
    new_starargs = fn(self.starargs) if self.starargs else None
    return ActualArgs(new_pos, new_keywords, new_starargs)

  def __eq__(self, other):
    if self.positional != other.positional or \
       len(self.keywords) != len(other.keywords) or \
       self.starargs != other.starargs:
      return False
    for (k,v) in self.keywords.iteritems():
      if k not in other.keywords:
        return False
      if other.keywords[k] != v:
        return False
    return True

  def __str__(self):
    arg_strings = []
    for p in self.positional:
      arg_strings.append(str(p))
    for (k,v) in self.keywords.items():
      arg_strings.append("%s = %s" % (k,v))
    if self.starargs:
      arg_strings.append("*" + str(self.starargs))
    return ", ".join(arg_strings)

  def __repr__(self):
    return str(self)

  def __hash__(self):
    if self._hash is None:
      self._hash = hash(self.positional) + len(self.keywords) + hash(self.starargs)
    return self._hash 
  
  def __iter__(self):
    return combine_iters(iter(self.positional),
                         self.keywords.itervalues(),
                         maybe_iter(self.starargs))

  def prepend_positional(self, more_args):
    new_pos = tuple(more_args) + self.positional
    return ActualArgs(new_pos, self.keywords, self.starargs)





from .. import prims 
from ..syntax import Const 
from ..analysis.value_range_analysis import Interval  
from range_transform import RangeTransform 

class RangePropagation(RangeTransform):
  
  def visit_Var(self, expr):
    if expr.name in self.ranges:
      range_val = self.ranges[expr.name]
      if isinstance(range_val, Interval) and range_val.lower == range_val.upper:
        return Const(range_val.lower, type = expr.type)
    return expr 
  
  def visit_PrimCall(self, expr):
    p = expr.prim 
    result = None 
    if isinstance(p, prims.Cmp):
      argx, argy = expr.args
      x, y = self.get(argx), self.get(argy)
      if x is not None and y is not None:        
        if p == prims.equal:
          result = self.cmp_eq(x,y)
        elif p == prims.less:
          result = self.cmp_lt(x,y)
        elif p == prims.less_equal:
          result = self.cmp_lte(x, y)
        elif p == prims.greater:
          result = self.cmp_lt(y, x)
        elif p == prims.greater_equal:
          result = self.cmp_lte(y, x)
     
          
    if result is None:
      return expr 
    else:
      return result 

  



from ..ndtypes import TupleT, ScalarT
from ..syntax import Expr 
from ..syntax.adverb_helpers import max_rank_arg, max_rank, unwrap_constant 
from ..syntax.helpers import unwrap_constant, wrap_if_constant 
 
from loop_builder import LoopBuilder

class AdverbBuilder(LoopBuilder):

  def list_none_axes(self, args, axis):
    axes = self.normalize_expr_axes(args, axis)
    return [self.is_none(axis) for axis in axes]
  
  def normalize_expr_axes(self, args, axis):
    """
    Given a list of arguments to an adverb and some value 
    or expression representing the axis/axes of iteration, 
    return a tuple of axis expressions, one per argument
    """
    axis = wrap_if_constant(axis)
    assert isinstance(axis, Expr), "Unexpected axis %s" % axis
    
    if isinstance(axis.type, TupleT):
      axes = self.tuple_elts(axis)
    else:
      assert isinstance(axis.type, ScalarT), "Unexpected axis %s : %s" % (axis, axis.type)
      axes = (axis,) * len(args)
    
    assert len(axes) == len(args), "Wrong number of axes (%d) for %d args" % (len(axes), len(args))
    
    if self.rank(max_rank_arg(args)) < 2:
        # if we don't actually have any multidimensional arguments, 
        # might as well make the axes just 0  
      axes = tuple(self.int(0) if self.is_none(axis) else axis for axis in axes)
    return axes 
    
  def normalize_axes(self, args, axis):
    """
    Given a list of arguments to an adverb and some value 
    or expression representing the axis/axes of iteration, 
    return a tuple of axis values, one per argument
    """

    if isinstance(axis, Expr) and isinstance(axis.type, TupleT):
      axis = self.tuple_elts(axis)
          
      
    # unpack the axis argument into a tuple,  
    # if only one axis was given, then repeat it as many times as we have args 
    if isinstance(axis, (list, tuple)):
      axes = tuple([unwrap_constant(elt) for elt in axis])
    elif isinstance(axis, Expr):
      axes = (unwrap_constant(axis),) * len(args) 
    else:
      assert axis is None or isinstance(axis, (int,long)), "Invalid axis %s" % axis
      axes = (axis,) * len(args)
    
    assert len(axes) == len(args), "Wrong number of axes (%d) for %d args" % (len(axes), len(args))
    
    if self.rank(max_rank_arg(args)) < 2:
        # if we don't actually have any multidimensional arguments, 
        # might as well make the axes just 0  
      axes = tuple(0 if axis is None else axis for axis in axes)
    return axes 
  
  def iter_bounds(self, args, axes, cartesian_product = False):
    """
    Given the arguments to an adverb (higher order array operator) and 
    the axes of the adverb's iteration, 
    given back a scalar or tuple representing the iteration space bounds
    the adverb will traverse
    """
    axes = self.normalize_axes(args, axes)
    assert len(args) == len(axes), "Mismatch between args %s and axes %s" % (args, axes) 
 
    
    if cartesian_product:
      bounds = []
      for arg, axis in zip(args, axes):
        if self.rank(arg) == 0:
          continue 
        if axis is None:
          bounds.extend(self.tuple_elts(self.shape(arg)))
        else:
          bounds.append(self.shape(arg, axis))
      return self.tuple(bounds)
    
    #...otherwise, we're doing an elementwise operation across
    #possibly multiple arguments 
    if any(axis is None for axis in axes):
      # if any of the axes are None then just find the highest rank argument 
      # which is going to be fully traversed and use its shape as the bounds
      # for the generated ParFor
      best_rank = -1 
      best_arg = None 
      for curr_arg, curr_axis in zip(args,axes):
        r = self.rank(curr_arg)
        if curr_axis is None and r > best_rank:
          best_rank = r
          best_arg = curr_arg 
      return self.shape(best_arg) 
       
    else:
      # if all axes are integer values, then keep the one with highest rank, 
      # it's bad that we're not doing any error checking here to make sure that 
      # all the non-scalar arguments have compatible shapes 
      best_rank = -1  
      best_arg = None
      best_axis = None 
      for curr_arg, curr_axis in zip(args,axes):
        r = self.rank(curr_arg)
        if r > best_rank:
          best_arg = curr_arg 
          best_axis = curr_axis
          best_rank = r 
      return self.shape(best_arg, best_axis)


import parakeet
import parakeet.testing_helpers
import numpy as np 


def matmult_high_level(X,Y):
  return np.array([[np.min(x+y) for y in Y.T] for x in X])

def test_matmult_tropical():
  n, d = 4,5
  m = 3
  X = np.random.randn(m,d)
  Y = np.random.randn(d,n)
  parakeet.testing_helpers.expect(matmult_high_level, [X,Y], matmult_high_level(X,Y))

if __name__ == '__main__':
    parakeet.testing_helpers.run_local_tests()





from dsltools import Node 
from expr import Expr
  


class Adverb(Expr, Node):
  _members = ['fn', 'output', 'type', 'source_info']
  
  def node_init(self):
    assert self.fn, "Can't construct adverb %s without a function argument" % self 
  def functions(self):
    yield self.fn 
    

      
class Accumulative(Expr):
  """
  Adverbs such as Reduce and Scan which carry an accumulated value and require a
  'combine' function to merge the accumulators resulting from parallel
  sub-computations.
  """
  _members = ['combine', 'init']
  
class HasEmit(Expr):
  """
  Common base class for Scan, IndexScan, and whatever other sorts of scans can be dreamed up
  """
  _members = ['emit']
  
  

class IndexAdverb(Adverb):
  _members = ['shape', 'start_index']
  
  
class IndexMap(IndexAdverb):
  """
  Map from each distinct index in the shape to a value 
  """
  pass 


class IndexAccumulative(IndexAdverb, Accumulative):
  pass 
  
class IndexReduce(IndexAccumulative):
  """
  Expect the 'fn' field to take indices and produces 
  element values, whereas 'combine' takes pairs of element 
  values and combines them. 
  """
  pass 

    
class IndexScan(IndexAccumulative, HasEmit):
  pass 
  
class DataAdverb(Adverb):
  _members = ['args', 'axis']

  def fn_to_str(self, fn):
    if hasattr(fn, 'name'):
      return fn.name
    else:
      return str(fn)

  def args_to_str(self):
    if isinstance(self.args, (list, tuple)):
      return ", ".join([str(arg) + ":" + str(arg.type) for arg in self.args])
    else:
      return str(self.args)

  def __repr__(self):
    s = "%s(axis = %s, args = (%s), type=%s, fn = %s)" % \
          (self.node_type(), self.axis,
           self.args_to_str(),
           self.type,
           self.fn_to_str(self.fn),)
    return s

  def __str__(self):
    return repr(self)

class Map(DataAdverb):
  pass 

class OuterMap(DataAdverb):
  pass 



class DataAccumulative(DataAdverb, Accumulative):
  pass 


class Reduce(DataAccumulative):
  def __repr__(self):
    return "Reduce(axis = %s, args = (%s), type = %s, init = %s, map_fn = %s, combine = %s)" % \
        (self.axis,
         self.args_to_str(),
         self.type,
         self.init,
         self.fn_to_str(self.fn),
         self.fn_to_str(self.combine))

class Scan(DataAccumulative, HasEmit):
  def __repr__(self):
    s = "%s(axis = %s, args = {%s}, type = %s, " \
        % (self.node_type(), self.axis,
           self.args_to_str(),
           self.type
          )
    s += "init = %s, map_fn = %s, combine = %s, emit = %s)" % \
         (self.init,
          self.fn_to_str(self.fn),
          self.fn_to_str(self.combine),
          self.fn_to_str(self.emit))
    return s




class Filter(Adverb):
  """
  Filters its arguments using the boolean predicate field 'fn'
  """
  pass 

class HasPred(Filter):
  _members = ['pred']


class Where(DataAdverb):
  """
  Applies 'fn' to each element of the arguments and 
  returns indices where 'pred' is True for the resulting
  element values
  """
  _members = ['pred']

class IndexFilter(IndexAdverb, Filter):
  pass 


class FilterReduce(Reduce, Filter):
  """
  Like a normal reduce but skips some elements if they don't pass
  the predicate 'pred'
  """
  pass  
  
  
class IndexFilterReduce(FilterReduce):
  pass 

class Tiled(object):
  _members = ['axes', 'fixed_tile_size']

  def __repr__(self):
    s = "%s(axes = %s, args = (%s), type=%s, fn = %s)" % \
          (self.node_type(), self.axes,
           self.args_to_str(),
           self.type,
           self.fn_to_str(self.fn),)
    return s

class TiledMap(Tiled, Map):
  pass

class TiledOuterMap(Tiled, OuterMap):
  pass

class TiledReduce(Tiled, Reduce):
  def __repr__(self):
    s = ("%s(axes = %s, args = {%s}, type = %s, init = %s, map_fn = %s, combine = %s)") % \
        (self.node_type(), self.axes,
         self.args_to_str(),
         self.type,
         self.init,
         self.fn_to_str(self.fn),
         self.fn_to_str(self.combine),
        )
    return s

class TiledScan(Tiled, Scan):
  def __repr__(self):
    s = "%s(axes = %s, args = {%s}, type = %s, " \
        % (self.node_type(), self.axes,
           self.args_to_str(),
           self.type
          )
    s += "init = %s, map_fn = %s, combine = %s, emit = %s)" % \
         (self.init,
          self.fn_to_str(self.fn),
          self.fn_to_str(self.combine),
          self.fn_to_str(self.emit))
    return s

  
class Conv(Adverb):
  _members = ['x', 'window_shape']
  
  def __repr__(self):
    return "Conv(fn = %s, x = %s, window_shape=%s)" % \
      (self.fn, self.x, self.window_shape)
      
  def __str__(self):
      return repr(self)

class ConvBorderFn(Conv):
  _members = ['border_fn']
  
class ConvBorderValue(Conv):
  _members = ['border_value']
  
class ConvPadding(Conv):
  _members = ['fill_value']
  
  


from ..analysis.value_range_analysis import (ValueRangeAnalyis, Interval, NoneValue)
from ..syntax.helpers import true, false 
from transform import Transform

class RangeTransform(Transform, ValueRangeAnalyis):
  """
  Base class for transforms which use value ranges gathered from analysis
  """
  
  def pre_apply(self, old_fn):
    ValueRangeAnalyis.__init__(self)
    self.visit_fn(old_fn)
    
  def cmp_eq(self, x, y):
    if not isinstance(x, Interval) or not isinstance(y, Interval):
      return None 
    if x.lower == y.lower and x.upper == y.upper:
      return true 
    if x.upper < y.lower or y.upper < x.lower:
      return false
    return None   
  
  def cmp_lt(self, x, y):
    if not isinstance(x, Interval) or not isinstance(y, Interval):
      return None 
    if x.upper < y.lower:
      return true 
    elif x.lower > y.upper:
      return false
    else:
      return None 
  
  def cmp_lte(self, x, y):
    if not isinstance(x, Interval) or not isinstance(y, Interval):
      return None 
    if x.upper < y.lower:
      return true 
    elif x.lower > y.upper:
      return false
    else:
      return None 
    